import os
import requests
import datetime

from typing import TypedDict, List, Dict, Any, cast

from semantic_version import Version
from pibble.util.files import load_yaml, load_json

__all__ = [
    "VersionDict",
    "check_make_directory",
    "get_local_installation_directory",
    "get_local_config_directory",
    "get_local_static_directory",
    "get_local_configuration",
    "get_version",
    "get_versions",
    "get_pending_versions",
]


class VersionDict(TypedDict):
    """
    The version dictionary.
    """

    version: Version
    release: datetime.date
    description: str


def get_local_installation_directory() -> str:
    """
    Gets where the local installation directory is (i.e. where the package data files are,
    either in ../site-packages/enfugue on a python installation, or ${ROOT}/enfugue in a
    precompiled package
    """
    here = os.path.dirname(os.path.abspath(__file__))
    while not os.path.basename(here) == "enfugue":
        here = os.path.abspath(os.path.join(here, "../"))
        if here == "/":
            raise IOError("Couldn't find installation directory.")
    return here


def get_local_config_directory() -> str:
    """
    Gets where the local configuration directory is.
    """
    here = os.path.dirname(os.path.abspath(__file__))
    while not os.path.isdir(os.path.join(here, "config")):
        here = os.path.abspath(os.path.join(here, "../"))
        if here == "/":
            raise IOError("Couldn't find config directory.")
    return os.path.join(here, "config")


def get_local_static_directory() -> str:
    """
    Gets where the local static directory is.
    """
    here = os.path.dirname(os.path.abspath(__file__))
    while not os.path.isdir(os.path.join(here, "static")):
        here = os.path.abspath(os.path.join(here, "../"))
        if here == "/":
            raise IOError("Couldn't find static directory.")
    return os.path.join(here, "static")


def check_make_directory(directory: str) -> None:
    """
    Checks if a directory doesn't exist, and makes it.
    Attempts to be thread-safe.
    """
    if not os.path.exists(directory):
        try:
            os.makedirs(directory)
            return
        except:
            if not os.path.exists(directory):
                raise
            return


def get_local_configuration() -> Dict[str, Any]:
    """
    Gets configuration from a file in the environment, or the base config.
    """
    default_config = os.path.join(get_local_config_directory(), "server.yml")
    config_file = os.getenv("ENFUGUE_CONFIG", default_config)
    if not os.path.exists(config_file):
        raise IOError(f"Configuration file {config_file} missing or inaccessible")

    basename, ext = os.path.splitext(os.path.basename(config_file))
    configuration: Dict[str, Any] = {}
    if ext.lower() in [".yml", ".yaml"]:
        configuration = load_yaml(config_file)
    elif ext.lower() == ".json":
        configuration = load_json(config_file)
    else:
        raise IOError(f"Unknown extension {ext}")
    if "configuration" in configuration:
        configuration = configuration["configuration"]
    return configuration


def get_version() -> Version:
    """
    Gets the version of enfugue installed.
    """
    from importlib.metadata import version, PackageNotFoundError

    try:
        return Version(version("enfugue"))
    except PackageNotFoundError:
        return "development"


def get_versions() -> List[VersionDict]:
    """
    Gets all version details from the CDN.
    """
    version_data = requests.get("https://cdn.enfugue.ai/versions.json").json()
    versions: List[VersionDict] = [
        cast(
            VersionDict,
            {
                "version": Version(datum["version"]),
                "release": datetime.datetime.strptime(datum["release"], "%Y-%m-%d").date(),
                "description": datum["description"],
            },
        )
        for datum in version_data
    ]
    versions.sort(key=lambda v: v["version"])
    return versions


def get_pending_versions() -> List[VersionDict]:
    """
    Gets only versions yet to be installed.
    """
    current_version = get_version()
    return [version for version in get_versions() if version["version"] > current_version]



import os
import glob
import PIL
import PIL.Image

from typing import Dict, List, Any
from webob import Request, Response

from pibble.ext.user.server.base import (
    UserExtensionServerBase,
    UserExtensionHandlerRegistry,
    UserExtensionTemplateServer,
)
from pibble.ext.session.server.base import SessionExtensionServerBase
from pibble.ext.rest.server.user import UserRESTExtensionServerBase
from pibble.api.middleware.database.orm import ORMMiddlewareBase
from pibble.api.exceptions import NotFoundError

from enfugue.diffusion.plan import DiffusionPlan
from enfugue.api.controller.base import EnfugueAPIControllerBase

__all__ = ["EnfugueAPIInvocationController"]


class EnfugueAPIInvocationController(EnfugueAPIControllerBase):
    handlers = UserExtensionHandlerRegistry()

    @property
    def thumbnail_height(self) -> int:
        """
        Gets the height of thumbnails.
        """
        return self.configuration.get("enfugue.thumbnail", 200)

    @handlers.path("^/api/invoke$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured("DiffusionInvocation", "create")
    def invoke_engine(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        Invokes the engine. Form a plan from the payload, then put the
        invocation in the queue.
        """
        # Get details about model
        model, size, lora, inversion, model_prompt, model_negative_prompt = (
            None,
            None,
            None,
            None,
            None,
            None,
        )
        model_name = request.parsed.pop("model", None)
        plan_kwargs: Dict[str, Any] = {}
        if model_name is not None:
            plan_kwargs = self.get_plan_kwargs_from_model(model_name)

        plan = DiffusionPlan.from_nodes(**{**plan_kwargs, **request.parsed})
        return self.invoke(request.token.user.id, plan).format()

    @handlers.path("^/api/invocation$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured("DiffusionInvocation", "read")
    def invocations(self, request: Request, response: Response) -> List[Dict[str, Any]]:
        """
        Gets all invocations since engine start for a user.
        """
        return [
            invocation.format()
            for invocation in self.manager.get_invocations(request.token.user.id)
        ]

    @handlers.path("^/api/invocation/intermediates/(?P<file_path>.+)$")
    @handlers.download()
    @handlers.methods("GET")
    @handlers.compress()
    @handlers.reverse("Intermediate", "/api/invocation/intermediates/{file_path}")
    @handlers.bypass(
        UserRESTExtensionServerBase,
        UserExtensionServerBase,
        ORMMiddlewareBase,
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
    )  # bypass processing for speed
    def download_intermediate_image(
        self, request: Request, response: Response, file_path: str
    ) -> str:
        """
        Downloads one of the intermediate results of an invocation.
        """
        image_path = os.path.join(self.manager.engine_intermediate_dir, file_path)
        if not os.path.exists(image_path):
            raise NotFoundError(f"No image at {file_path}")
        return image_path

    @handlers.path("^/api/invocation/images/(?P<file_path>.+)$")
    @handlers.download()
    @handlers.methods("GET")
    @handlers.compress()
    @handlers.cache()
    @handlers.reverse("Image", "/api/invocation/images/{file_path}")
    @handlers.bypass(
        UserRESTExtensionServerBase,
        UserExtensionServerBase,
        ORMMiddlewareBase,
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
    )  # bypass processing for speed
    def download_image(self, request: Request, response: Response, file_path: str) -> str:
        """
        Downloads one of the results of an invocation.
        """
        image_path = os.path.join(self.manager.engine_image_dir, file_path)
        if not os.path.exists(image_path):
            raise NotFoundError(f"No image at {file_path}")
        return image_path

    @handlers.path("^/api/invocation/thumbnails/(?P<file_path>.+)$")
    @handlers.download()
    @handlers.methods("GET")
    @handlers.compress()
    @handlers.cache()
    @handlers.reverse("Image", "/api/invocation/thumbnails/{file_path}")
    @handlers.bypass(
        UserRESTExtensionServerBase,
        UserExtensionServerBase,
        ORMMiddlewareBase,
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
    )  # bypass processing for speed
    def download_image_thumbnail(self, request: Request, response: Response, file_path: str) -> str:
        """
        Downloads one of the results of an invocation only as a thumbnail
        """
        image_path = os.path.join(self.manager.engine_image_dir, file_path)
        if not os.path.exists(image_path):
            raise NotFoundError(f"No image at {file_path}")
        image_name, ext = os.path.splitext(os.path.basename(image_path))
        thumbnail_path = os.path.join(self.manager.engine_image_dir, f"{image_name}_thumb{ext}")
        if not os.path.exists(thumbnail_path):
            image = PIL.Image.open(image_path)
            width, height = image.size
            scale = self.thumbnail_height / height
            image.resize((int(width * scale), int(height * scale))).save(thumbnail_path)
        return thumbnail_path

    @handlers.path("^/api/invocation/nsfw$")
    @handlers.methods("GET")
    @handlers.cache()
    @handlers.reverse("NSFW", "/api/invocation/nsfw")
    @handlers.bypass(
        UserRESTExtensionServerBase,
        UserExtensionServerBase,
        ORMMiddlewareBase,
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
    )  # bypass processing for speed
    def download_nsfw_image(self, request: Request, response: Response) -> None:
        """
        The image served when NSFW content was detected
        """
        contents = '<svg viewBox="0 0 512 512" xmlns="http://www.w3.org/2000/svg" style="background-color: #222222;" width="512" height="512">'
        contents += "<style>text { font: bold 25px sans-serif; fill: #555555; letter-spacing: 1px; } g { transform: rotateZ(-45deg); transform-origin: center; }</style><g>"
        for i in range(40):
            for j in range(24):
                contents += f'<text x="{(j*80)+((i%10)*8)-140}" y="{i*24-80}">NSFW</text>'
        contents += "</g></svg>"
        response.content_type = "image/svg+xml"
        response.text = contents

    @handlers.path("^/api/invocation/(?P<uuid>[a-zA-Z0-9\-]+)$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured("DiffusionInvocation", "read")
    def invocation(self, request: Request, response: Response, uuid: str) -> Dict[str, Any]:
        """
        Gets a single set of invocation details by UUID.
        """
        for invocation in self.manager.get_invocations(request.token.user.id):
            if invocation.uuid == uuid:
                formatted = invocation.format()
                database_invocation = (
                    self.database.query(self.orm.DiffusionInvocation)
                    .filter(self.orm.DiffusionInvocation.id == uuid)
                    .one()
                )
                formatted_images = formatted.get("images", [])
                if isinstance(formatted_images, list):
                    database_invocation.outputs = len(formatted_images)
                database_invocation.duration = formatted.get("duration", 0)
                if "message" in formatted:
                    database_invocation.error = formatted["message"]
                self.database.commit()
                return formatted
        raise NotFoundError(f"No invocation matching UUID {uuid}")

    @handlers.path("^/api/invocation/(?P<uuid>[a-zA-Z0-9\-]+)$")
    @handlers.methods("DELETE")
    @handlers.format()
    @handlers.secured("DiffusionInvocation", "delete")
    def delete_invocation(self, request: Request, response: Response, uuid: str) -> None:
        """
        Deletes an invocation.
        """
        database_invocation = (
            self.database.query(self.orm.DiffusionInvocation)
            .filter(self.orm.DiffusionInvocation.id == uuid)
            .one_or_none()
        )
        if not database_invocation:
            raise NotFoundError(f"No invocation with ID {uuid}")

        for dirname in [self.manager.engine_image_dir, self.manager.engine_intermediate_dir]:
            for invocation_image in glob.glob(f"{uuid}*.png", root_dir=dirname):
                os.remove(os.path.join(dirname, invocation_image))
        self.database.delete(database_invocation)
        self.database.commit()

    @handlers.path("^/api/invocation/stop$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured()
    def stop_engine(self, request: Request, response: Response) -> None:
        """
        Stops the engine and any invocations.
        """
        self.manager.stop_engine()



from __future__ import annotations

import os
import gc
import onnx
import torch
import logging
import tensorrt as trt

from copy import copy
from typing import Dict, Optional, Tuple, Any, List

from polygraphy import cuda
from collections import OrderedDict

from polygraphy.logger import G_LOGGER
from polygraphy.backend.common import bytes_from_path
from polygraphy.backend.trt import CreateConfig as TRTCreateConfig
from polygraphy.backend.trt import Profile as TRTProfile
from polygraphy.backend.trt import util as trt_util
from polygraphy.backend.trt import (
    engine_from_bytes,
    engine_from_network,
    network_from_onnx_path,
    save_engine,
)

from enfugue.util import logger
from enfugue.diffusion.util import DTypeConverter
from enfugue.diffusion.rt.model import BaseModel


class Engine:
    """
    This class manages state for TensorRT engines.

    Building a TRT engine is a three-step process:
    1. Generate ONNX model from NN module
    2. Optimize ONNX model
    3. Compile TRT engine from optimized ONNX model
    """

    engine: trt.ICudaEngine
    context: trt.IExecutionContext
    buffers: OrderedDict[str, cuda.DeviceView]
    tensors: OrderedDict[str, torch.Tensor]

    def __init__(self, engine_path: str) -> None:
        """ """
        self.engine_path = engine_path
        self.engine = None
        self.context = None
        self.buffers = OrderedDict()
        self.tensors = OrderedDict()
        self.hijack_logger()

    def hijack_logger(self) -> None:
        """
        Sets the TRT logger to use the enfugue logger.
        """

        def _g_logger_log(msg: str, severity: int = logging.DEBUG, **kwargs: Any) -> None:
            if isinstance(msg, str) and msg.startswith("Total # of Profiles"):
                return  # Ignore this log spam
            if severity not in [
                logging.DEBUG,
                logging.INFO,
                logging.ERROR,
                logging.CRITICAL,
                logging.WARNING,
            ]:
                severity = logging.DEBUG
            logger.log(severity, msg)

        def _trt_logger_log(severity: trt.tensorrt.ILogger.Severity, msg: str) -> None:
            converted_severity = {
                trt.Logger.INTERNAL_ERROR: logging.CRITICAL,
                trt.Logger.ERROR: logging.ERROR,
                trt.Logger.WARNING: logging.WARNING,
                trt.Logger.INFO: logging.INFO,
                trt.Logger.VERBOSE: logging.DEBUG,
            }.get(severity, logging.DEBUG)
            logger.log(converted_severity, msg)

        G_LOGGER.log = _g_logger_log
        trt_util.get_trt_logger().log = _trt_logger_log

    def __del__(self) -> None:
        """
        Deletes the engine.
        """
        for buf in self.buffers.values():
            try:
                buf.free()
            except:
                pass
        del self.engine
        del self.context
        del self.buffers
        del self.tensors

    def load(self) -> None:
        """
        Loads the engine.
        """
        logger.debug(f"Loading TensorRT engine: {self.engine_path}")
        self.engine = engine_from_bytes(bytes_from_path(self.engine_path))

    def activate(self) -> None:
        """
        Creates execution context from the engine.
        """
        self.context = self.engine.create_execution_context()

    def allocate_buffers(self, shape_dict: Optional[Dict[str, Tuple]] = None, device: str = "cuda"):
        """
        Allocates TRT buffers.
        """
        bindings_per_profile = trt_util.get_bindings_per_profile(self.engine)

        for i in range(bindings_per_profile):
            binding = self.engine[i]

            if shape_dict and binding in shape_dict:
                shape = shape_dict[binding]
            else:
                shape = self.engine.get_binding_shape(binding)

            dtype = trt.nptype(self.engine.get_binding_dtype(binding))

            if self.engine.binding_is_input(binding):
                self.context.set_binding_shape(i, shape)

            tensor = torch.empty(tuple(shape), dtype=DTypeConverter.from_numpy(dtype)).to(
                device=device
            )
            self.tensors[binding] = tensor
            self.buffers[binding] = cuda.DeviceView(ptr=tensor.data_ptr(), shape=shape, dtype=dtype)
            logger.debug(f"Binding {binding} to tensor of shape {shape}")

    def infer(
        self, feeds: Dict[str, cuda.DeviceView], stream: cuda.Stream
    ) -> OrderedDict[str, torch.Tensor]:
        """
        Runs inference through the engine.
        """
        start_binding, end_binding = trt_util.get_active_profile_bindings(self.context)

        # Shallow copy of ordered dict
        device_buffers = copy(self.buffers)
        for name, buf in feeds.items():
            assert isinstance(buf, cuda.DeviceView)
            device_buffers[name] = buf

        bindings = [0] * start_binding + [buf.ptr for buf in device_buffers.values()]
        success = self.context.execute_async_v2(bindings=bindings, stream_handle=stream.ptr)
        if not success:
            raise IOError("Inference failed!")

        return self.tensors

    @staticmethod
    def get_onnx_path(engine_root: str, opt: bool = False) -> str:
        """
        Gets the path for the ONNX model in a directory.
        """
        model_name = f"model.opt" if opt else "model"
        return os.path.join(engine_root, f"{model_name}.onnx")

    @staticmethod
    def get_timing_cache(engine_root: str) -> str:
        """
        Gets the path for the timing cache in a directory.
        """
        return os.path.join(engine_root, "timing_cache")

    @staticmethod
    def get_engine_path(engine_root: str) -> str:
        """
        Gets the path for the engine in a directory.
        """
        return os.path.join(engine_root, "engine.plan")

    @staticmethod
    def get_paths(engine_root: str) -> Tuple[str, str, str, str]:
        """
        Gets the four paths; timing cache, engine path, model, and opt model.
        """
        return (
            Engine.get_timing_cache(engine_root),
            Engine.get_engine_path(engine_root),
            Engine.get_onnx_path(engine_root),
            Engine.get_onnx_path(engine_root, True),
        )

    def build(
        self,
        onnx_path: str,
        use_fp16: bool,
        input_profile: Optional[Dict[str, List[Tuple[int, ...]]]] = None,
        enable_preview: bool = False,
        enable_all_tactics: bool = False,
        timing_cache: str = "timing_cache",
        workspace_size: int = 0,
    ) -> None:
        """
        Builds a TRT engine.
        """
        dtype = "fp16" if use_fp16 else "fp32"
        logger.debug(f"Building TensorRT engine for {onnx_path}: {self.engine_path} ({dtype})")
        profile = TRTProfile()
        if input_profile:
            for name, dims in input_profile.items():
                profile.add(name, min=dims[0], opt=dims[1], max=dims[2])

        config_kwargs: dict[str, Any] = {}

        config_kwargs["preview_features"] = [
            trt.PreviewFeature.DISABLE_EXTERNAL_TACTIC_SOURCES_FOR_CORE_0805
        ]
        if enable_preview:
            # Faster dynamic shapes made optional since it increases engine build time.
            config_kwargs["preview_features"].append(trt.PreviewFeature.FASTER_DYNAMIC_SHAPES_0805)
        if workspace_size > 0:
            config_kwargs["memory_pool_limits"] = {trt.MemoryPoolType.WORKSPACE: workspace_size}
        if not enable_all_tactics:
            config_kwargs["tactic_sources"] = []

        config = TRTCreateConfig(
            fp16=use_fp16,
            profiles=[profile],
            load_timing_cache=timing_cache,
            **config_kwargs,
        )
        engine = engine_from_network(
            network_from_onnx_path(onnx_path),
            config=config,
            save_timing_cache=timing_cache,
        )
        save_engine(engine, self.engine_path)

    @staticmethod
    def build_all(
        models: Dict[str, BaseModel],
        onnx_opset: int,
        opt_image_height: int,
        opt_image_width: int,
        opt_batch_size: int = 1,
        use_fp16: bool = False,
        force_engine_rebuild: bool = False,
        static_batch: bool = False,
        static_shape: bool = True,
        enable_preview: bool = False,
        enable_all_tactics: bool = False,
        max_workspace_size: int = 0,
    ) -> Dict[str, Engine]:
        """
        Builds all TRT engines at once.
        """
        built_engines = {}

        # Export models to ONNX
        for model_dir, model_obj in models.items():
            model_name = model_obj.get_model_key()
            timing_cache, engine_path, onnx_path, onnx_opt_path = Engine.get_paths(model_dir)
            if force_engine_rebuild or not os.path.exists(engine_path):
                logger.debug(f"Building engine for {model_name} at {model_dir}")
                if force_engine_rebuild or not os.path.exists(onnx_opt_path):
                    if force_engine_rebuild or not os.path.exists(onnx_path):
                        logger.debug(f"Exporting model to {onnx_path}")
                        model = model_obj.get_model()
                        with torch.inference_mode(), torch.autocast("cuda"):
                            inputs = model_obj.get_sample_input(
                                opt_batch_size, opt_image_height, opt_image_width
                            )
                            torch.onnx.export(
                                model,
                                inputs,
                                onnx_path,
                                export_params=True,
                                opset_version=onnx_opset,
                                do_constant_folding=True,
                                input_names=model_obj.get_input_names(),
                                output_names=model_obj.get_output_names(),
                                dynamic_axes=model_obj.get_dynamic_axes(),
                            )
                        del model
                        torch.cuda.empty_cache()
                        gc.collect()
                    else:
                        logger.debug(f"Found cached model at {onnx_path}")

                    # Optimize onnx
                    if force_engine_rebuild or not os.path.exists(onnx_opt_path):
                        logger.debug(f"Generating optimized model to {onnx_opt_path}")
                        onnx_opt_graph = model_obj.optimize(onnx.load(onnx_path))
                        onnx.save_model(onnx_opt_graph, onnx_opt_path)
                    else:
                        logger.debug(f"Found cached optimized model at {onnx_opt_path}")

        # Build TensorRT engines
        for model_dir, model_obj in models.items():
            model_name = model_obj.get_model_key()
            timing_cache, engine_path, onnx_path, onnx_opt_path = Engine.get_paths(model_dir)
            engine = Engine(engine_path)
            logger.debug(
                model_obj.get_input_profile(
                    opt_batch_size,
                    opt_image_height,
                    opt_image_width,
                    static_batch=static_batch,
                    static_shape=static_shape,
                )
            )

            if force_engine_rebuild or not os.path.exists(engine.engine_path):
                logger.debug(f"Building TensorRT engine for {model_name}")
                engine.build(
                    onnx_opt_path,
                    use_fp16=use_fp16,
                    input_profile=model_obj.get_input_profile(
                        opt_batch_size,
                        opt_image_height,
                        opt_image_width,
                        static_batch=static_batch,
                        static_shape=static_shape,
                    ),
                    enable_preview=enable_preview,
                    timing_cache=timing_cache,
                    workspace_size=max_workspace_size,
                )
            built_engines[model_name] = engine

        # Load and activate TensorRT engines
        for model_dir, model_obj in models.items():
            model_name = model_obj.get_model_key()
            engine = built_engines[model_name]
            engine.load()
            engine.activate()

        return built_engines



# type: ignore
# Taken from https://github.com/TencentARC/GFPGAN/blob/master/gfpgan/utils.py
# Slightly adapted to allow passing face restore model root path
import cv2
import os
import torch
from basicsr.utils import img2tensor, tensor2img
from basicsr.utils.download_util import load_file_from_url
from facexlib.utils.face_restoration_helper import FaceRestoreHelper
from torchvision.transforms.functional import normalize

from gfpgan.archs.gfpgan_bilinear_arch import GFPGANBilinear
from gfpgan.archs.gfpganv1_arch import GFPGANv1
from gfpgan.archs.gfpganv1_clean_arch import GFPGANv1Clean

ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class GFPGANer:
    """Helper for restoration with GFPGAN.

    It will detect and crop faces, and then resize the faces to 512x512.
    GFPGAN is used to restored the resized faces.
    The background is upsampled with the bg_upsampler.
    Finally, the faces will be pasted back to the upsample background image.

    Args:
        model_path (str): The path to the GFPGAN model. It can be urls (will first download it automatically).
        upscale (float): The upscale of the final output. Default: 2.
        arch (str): The GFPGAN architecture. Option: clean | original. Default: clean.
        channel_multiplier (int): Channel multiplier for large networks of StyleGAN2. Default: 2.
        bg_upsampler (nn.Module): The upsampler for the background. Default: None.
    """

    def __init__(
        self,
        model_path,
        upscale=2,
        arch="clean",
        channel_multiplier=2,
        bg_upsampler=None,
        device=None,
        rootpath=None,
    ):
        self.upscale = upscale
        self.bg_upsampler = bg_upsampler

        # initialize model
        self.device = (
            torch.device("cuda" if torch.cuda.is_available() else "cpu")
            if device is None
            else device
        )
        # initialize the GFP-GAN
        if arch == "clean":
            self.gfpgan = GFPGANv1Clean(
                out_size=512,
                num_style_feat=512,
                channel_multiplier=channel_multiplier,
                decoder_load_path=None,
                fix_decoder=False,
                num_mlp=8,
                input_is_latent=True,
                different_w=True,
                narrow=1,
                sft_half=True,
            )
        elif arch == "bilinear":
            self.gfpgan = GFPGANBilinear(
                out_size=512,
                num_style_feat=512,
                channel_multiplier=channel_multiplier,
                decoder_load_path=None,
                fix_decoder=False,
                num_mlp=8,
                input_is_latent=True,
                different_w=True,
                narrow=1,
                sft_half=True,
            )
        elif arch == "original":
            self.gfpgan = GFPGANv1(
                out_size=512,
                num_style_feat=512,
                channel_multiplier=channel_multiplier,
                decoder_load_path=None,
                fix_decoder=True,
                num_mlp=8,
                input_is_latent=True,
                different_w=True,
                narrow=1,
                sft_half=True,
            )
        elif arch == "RestoreFormer":
            from gfpgan.archs.restoreformer_arch import RestoreFormer

            self.gfpgan = RestoreFormer()
        # initialize face helper
        self.face_helper = FaceRestoreHelper(
            upscale,
            face_size=512,
            crop_ratio=(1, 1),
            det_model="retinaface_resnet50",
            save_ext="png",
            use_parse=True,
            device=self.device,
            model_rootpath="gfpgan/weights" if not rootpath else rootpath,
        )

        if model_path.startswith("https://"):
            model_path = load_file_from_url(
                url=model_path,
                model_dir=os.path.join(ROOT_DIR, "gfpgan/weights"),
                progress=True,
                file_name=None,
            )
        loadnet = torch.load(model_path)
        if "params_ema" in loadnet:
            keyname = "params_ema"
        else:
            keyname = "params"
        self.gfpgan.load_state_dict(loadnet[keyname], strict=True)
        self.gfpgan.eval()
        self.gfpgan = self.gfpgan.to(self.device)

    @torch.no_grad()
    def enhance(self, img, has_aligned=False, only_center_face=False, paste_back=True, weight=0.5):
        self.face_helper.clean_all()

        if has_aligned:  # the inputs are already aligned
            img = cv2.resize(img, (512, 512))
            self.face_helper.cropped_faces = [img]
        else:
            self.face_helper.read_image(img)
            # get face landmarks for each face
            self.face_helper.get_face_landmarks_5(
                only_center_face=only_center_face, eye_dist_threshold=5
            )
            # eye_dist_threshold=5: skip faces whose eye distance is smaller than 5 pixels
            # TODO: even with eye_dist_threshold, it will still introduce wrong detections and restorations.
            # align and warp each face
            self.face_helper.align_warp_face()

        # face restoration
        for cropped_face in self.face_helper.cropped_faces:
            # prepare data
            cropped_face_t = img2tensor(cropped_face / 255.0, bgr2rgb=True, float32=True)
            normalize(cropped_face_t, (0.5, 0.5, 0.5), (0.5, 0.5, 0.5), inplace=True)
            cropped_face_t = cropped_face_t.unsqueeze(0).to(self.device)

            try:
                output = self.gfpgan(cropped_face_t, return_rgb=False, weight=weight)[0]
                # convert to image
                restored_face = tensor2img(output.squeeze(0), rgb2bgr=True, min_max=(-1, 1))
            except RuntimeError as error:
                print(f"\tFailed inference for GFPGAN: {error}.")
                restored_face = cropped_face

            restored_face = restored_face.astype("uint8")
            self.face_helper.add_restored_face(restored_face)

        if not has_aligned and paste_back:
            # upsample the background
            if self.bg_upsampler is not None:
                # Now only support RealESRGAN for upsampling background
                bg_img = self.bg_upsampler.enhance(img, outscale=self.upscale)[0]
            else:
                bg_img = None

            self.face_helper.get_inverse_affine(None)
            # paste each restored face to the input image
            restored_img = self.face_helper.paste_faces_to_input_image(upsample_img=bg_img)
            return self.face_helper.cropped_faces, self.face_helper.restored_faces, restored_img
        else:
            return self.face_helper.cropped_faces, self.face_helper.restored_faces, None



from typing import TypedDict
from enfugue.util.gputil import getGPUs  # type: ignore[attr-defined]

__all__ = ["GPUMemoryStatusDict", "GPUStatusDict", "get_gpu_status"]


class GPUMemoryStatusDict(TypedDict):
    """
    The memory status dictionary.
    """

    free: int
    total: int
    used: int
    util: float


class GPUStatusDict(TypedDict):
    """
    The GPU status dictionary.
    """

    driver: str
    name: str
    load: float
    temp: float
    memory: GPUMemoryStatusDict


def get_gpu_status() -> GPUStatusDict:
    """
    Gets current GPU status.
    """

    primary_gpu = getGPUs()[0]

    return {
        "driver": primary_gpu.driver,
        "name": primary_gpu.name,
        "load": primary_gpu.load,
        "temp": primary_gpu.temperature,
        "memory": {
            "free": primary_gpu.memoryFree,
            "total": primary_gpu.memoryTotal,
            "used": primary_gpu.memoryUsed,
            "util": primary_gpu.memoryUtil,
        },
    }



import re
import datetime

from typing import Any, Optional, List, Dict
from pibble.util.helpers import resolve
from pibble.util.strings import get_uuid

from enfugue.util import logger
from enfugue.diffusion.engine import DiffusionEngine
from enfugue.diffusion.plan import DiffusionPlan
from multiprocessing import Lock

__all__ = ["Invocation"]


class TerminatedError(Exception):
    pass


class Invocation:
    """
    Holds the details for a single invocation
    """

    start_time: Optional[datetime.datetime]
    end_time: Optional[datetime.datetime]
    last_intermediate_time: Optional[datetime.datetime]
    results: Optional[List[str]]
    last_images: Optional[List[str]]
    last_step: Optional[int]
    last_total: Optional[int]
    last_rate: Optional[float]
    id: Optional[int]
    error: Optional[Exception]

    def __init__(
        self,
        engine: DiffusionEngine,
        plan: DiffusionPlan,
        engine_image_dir: str,
        engine_intermediate_dir: str,
        decode_nth_intermediate: Optional[int] = 10,
        communication_timeout: Optional[int] = 90,  # Really big images can take this long to start
        metadata: Optional[Dict[str, Any]] = None,
        save: bool = True,
        **kwargs: Any,
    ) -> None:
        self.lock = Lock()
        self.uuid = get_uuid()
        self.engine = engine
        self.plan = plan

        self.results_dir = engine_image_dir
        self.intermediate_dir = engine_intermediate_dir
        self.intermediate_steps = decode_nth_intermediate
        self.communication_timeout = communication_timeout
        self.metadata = metadata
        self.save = save

        self.id = None
        self.error = None
        self.start_time = None
        self.end_time = None
        self.last_intermediate_time = None
        self.last_step = None
        self.last_total = None
        self.last_rate = None
        self.last_images = None
        self.results = None

    def _communicate(self) -> None:
        """
        Tries to communicate with the engine to see what's going on.
        """
        if self.id is None:
            raise IOError("Invocation not started yet.")
        if self.results is not None:
            raise IOError("Invocation already completed.")
        try:
            last_intermediate = self.engine.last_intermediate(self.id)
            if last_intermediate is not None:
                for key in ["step", "total", "images", "rate"]:
                    if key in last_intermediate:
                        setattr(self, f"last_{key}", last_intermediate[key])
                self.last_intermediate_time = datetime.datetime.now()
            try:
                result = self.engine.wait(self.id, timeout=0.1)
            except TimeoutError:
                raise
            except Exception as ex:
                result = None
                self.error = ex
            if result is not None:
                # Complete
                self.results = []
                self.end_time = datetime.datetime.now()
                if "images" in result:
                    is_nsfw = result.get("nsfw_content_detected", [])
                    for i, image in enumerate(result["images"]):
                        if len(is_nsfw) > i and is_nsfw[i]:
                            self.results.append("nsfw")
                        elif self.save:
                            image_path = f"{self.results_dir}/{self.uuid}_{i}.png"
                            image.save(image_path)
                            self.results.append(image_path)
                        else:
                            self.results.append("unsaved")
                elif "error" in result:
                    error_type = resolve(result["error"])
                    self.error = error_type(result["message"])
                    if "traceback" in result:
                        logger.error(f"Traceback for invocation {self.uuid}:")
                        logger.debug(result["traceback"])
        except TimeoutError:
            return

    def _check_raise_error(self) -> None:
        """
        Raises an error if one has been set.
        """
        if self.error is not None:
            raise self.error

    def start(self) -> None:
        """
        Starts the invocation (locks)
        """
        with self.lock:
            self.start_time = datetime.datetime.now()
            payload = self.plan.get_serialization_dict()
            payload["intermediate_dir"] = self.intermediate_dir
            payload["intermediate_steps"] = self.intermediate_steps
            self.id = self.engine.dispatch("plan", payload)

    def poll(self) -> None:
        """
        Calls communicate once (locks)
        """
        with self.lock:
            self._communicate()

    @property
    def is_dangling(self) -> bool:
        """
        Determine if this invocation appears lost.
        """
        if self.id is None or self.start_time is None:
            return False
        if self.end_time is not None and self.results is not None or self.error is not None:
            return False
        if self.communication_timeout is None:
            return False
        if self.last_intermediate_time is not None:
            last_known_time = self.last_intermediate_time
        else:
            last_known_time = self.start_time
        seconds_since_last_communication = (
            datetime.datetime.now() - last_known_time
        ).total_seconds()
        return seconds_since_last_communication > self.communication_timeout

    def timeout(self) -> None:
        """
        Times out an invocation that got lost
        """
        self.error = TimeoutError("Invocation timed out.")
        self.end_time = datetime.datetime.now()

    def terminate(self) -> None:
        """
        Kills an active invocation
        """
        if self.id is None:
            raise IOError("Invocation not started yet.")
        if self.results is not None:
            raise IOError("Invocation completed.")
        if self.error is None:
            self.error = TerminatedError("The invocation was terminated prematurely")
        else:
            raise IOError(f"Invocation already ended in error {self.error}")

    def format(self) -> Dict[str, Any]:
        """
        Formats the invocation to a dictionary
        """
        with self.lock:
            if self.id is None or self.start_time is None:
                return {"status": "queued", "uuid": self.uuid}

            if self.error is not None:
                return {
                    "status": "error",
                    "uuid": self.uuid,
                    "message": str(self.error),
                }

            images = None
            if self.results is not None:
                status = "completed"
                images = ["/".join(re.split(r"/|\\", path)[-2:]) for path in self.results]
            else:
                status = "processing"
                self._communicate()
                if self.last_images is not None:
                    images = ["/".join(re.split(r"/|\\", path)[-2:]) for path in self.last_images]

            if self.end_time is None:
                duration = (datetime.datetime.now() - self.start_time).total_seconds()
            else:
                duration = (self.end_time - self.start_time).total_seconds()

            step, total, progress, rate = None, None, None, None
            if self.last_total is not None and self.last_total > 0:
                total = self.last_total
            if self.last_step is not None:
                step = total if self.results is not None else self.last_step
            if total is not None and step is not None:
                progress = step / total
            if self.last_rate is not None:
                rate = self.last_rate
            elif step is not None:
                rate = step / duration

            formatted = {
                "id": self.id,
                "uuid": self.uuid,
                "status": status,
                "progress": progress,
                "step": step,
                "duration": duration,
                "total": total,
                "images": images,
                "rate": rate,
            }
            if self.metadata:
                formatted["metadata"] = self.metadata
            return formatted

    def __str__(self) -> str:
        """
        Stringifies the invocation for debugging.
        """

        return f"Invocation {self.uuid}, last step: {self.last_step}, last total: {self.last_total}: error: {self.error}, results: {self.results}"



from enfugue.api.server import EnfugueAPIServer
from enfugue.api.manager import SystemManager
from enfugue.api.downloads import Download
from enfugue.api.invocations import Invocation

EnfugueAPIServer, Download, SystemManager, Invocation

__all__ = ["EnfugueAPIServer", "Download", "SystemManager", "Invocation"]



from __future__ import annotations

import cv2
import PIL
import numpy as np

from enfugue.util import check_download_to_dir
from enfugue.diffusion.vision import ComputerVision

__all__ = ["EdgeDetector"]


class EdgeDetector:
    """
    Provides edge detection methods
    """

    HED_PROTOTXT = "https://github.com/ashukid/hed-edge-detector/raw/master/deploy.prototxt"
    HED_CAFFEMODEL = (
        "https://github.com/ashukid/hed-edge-detector/raw/master/hed_pretrained_bsds.caffemodel"
    )
    HED_MEAN = (104.00698793, 116.66876762, 122.67891434)
    MLSD_WEIGHTS = "https://github.com/lhwcv/mlsd_pytorch/raw/main/models/mlsd_large_512_fp32.pth"

    def __init__(self, model_dir: str) -> None:
        """
        On initialization, pass the model dir.
        """
        self.model_dir = model_dir

    @property
    def hed_prototxt(self) -> str:
        """
        Gets the local path to the HED prototxt.
        """
        return check_download_to_dir(self.HED_PROTOTXT, self.model_dir)

    @property
    def hed_caffemodel(self) -> str:
        """
        Gets the local path to the HED caffemodel.
        """
        return check_download_to_dir(self.HED_CAFFEMODEL, self.model_dir)

    @property
    def mlsd_weights(self) -> str:
        """
        Gets the local path to the MLSD weights file.
        """
        return check_download_to_dir(self.MLSD_WEIGHTS, self.model_dir)

    @staticmethod
    def canny(image: PIL.Image.Image, lower: int = 100, upper: int = 200) -> PIL.Image.Image:
        """
        Runs canny edge detection on an image.
        """
        canny = cv2.Canny(np.array(image), lower, upper)[:, :, None]
        return PIL.Image.fromarray(np.concatenate([canny, canny, canny], axis=2))

    def hed(self, image: PIL.Image.Image) -> PIL.Image.Image:
        """
        Runs holistically-nested edge detection on an image.
        """
        from enfugue.diffusion.edge.hed import HEDCropLayer  # type: ignore[attr-defined]

        width, height = image.size

        model = cv2.dnn.readNetFromCaffe(self.hed_prototxt, self.hed_caffemodel)
        cv2.dnn_registerLayer("Crop", HEDCropLayer)

        cv2_image = ComputerVision.convert_image(image)
        dnn_input = cv2.dnn.blobFromImage(
            cv2_image,
            scalefactor=1.0,
            size=(width, height),
            mean=self.HED_MEAN,
            swapRB=False,
            crop=False,
        )
        model.setInput(dnn_input)
        dnn_output = model.forward()
        output_array = (cv2.cvtColor(dnn_output[0, 0], cv2.COLOR_GRAY2BGR) * 255).astype(np.uint8)
        cv2.dnn_unregisterLayer("Crop")
        return PIL.Image.fromarray(output_array)

    def mlsd(self, image: PIL.Image.Image) -> PIL.Image.Image:
        """
        Runs Mobile Line Segment Detection (MLSD) on an image.
        """
        import torch
        from enfugue.diffusion.edge.mlsd import MLSD, pred_lines  # type: ignore[attr-defined]

        if torch.cuda.is_available():
            model = MLSD().cuda().eval()
            device = torch.device("cuda")
        else:
            model = MLSD().eval()
            device = torch.device("cpu")

        model.load_state_dict(torch.load(self.mlsd_weights, map_location=device), strict=True)
        cv2_image = ComputerVision.convert_image(image)
        cv2_image = cv2.resize(cv2_image, (512, 512))
        cv2_image = cv2.cvtColor(cv2_image, cv2.COLOR_BGR2RGB)
        lines = pred_lines(cv2_image, model, [512, 512], 0.1, 20)
        cv2_image = np.zeros((512, 512, 3), np.uint8)

        for line in lines:
            cv2.line(
                cv2_image,
                (int(line[0]), int(line[1])),
                (int(line[2]), int(line[3])),
                (255, 255, 255),
                1,
                16,
            )
        # Unload model and wipe cache
        del model
        torch.cuda.empty_cache()
        return ComputerVision.revert_image(cv2_image).resize(image.size)



from typing import (
    Any,
    List,
    Callable,
    Dict,
    List,
    Optional,
    Union,
    Tuple,
    Iterator,
    Literal,
    Mapping,
    cast,
)

import PIL
import PIL.Image
import math
import torch
import safetensors.torch
import datetime
import numpy as np

from contextlib import contextmanager
from collections import defaultdict

from transformers import (
    CLIPImageProcessor,
    CLIPTextModel,
    CLIPTokenizer,
)
from diffusers.schedulers import KarrasDiffusionSchedulers, DDIMScheduler

from diffusers.models import AutoencoderKL, UNet2DConditionModel, ControlNetModel

from diffusers.pipelines.stable_diffusion import (
    StableDiffusionPipeline,
    StableDiffusionPipelineOutput,
    StableDiffusionSafetyChecker,
)

from diffusers.utils import randn_tensor, PIL_INTERPOLATION
from diffusers.image_processor import VaeImageProcessor

from enfugue.util import logger

# This is ~64k×64k. Absurd, but I don't judge
PIL.Image.MAX_IMAGE_PIXELS = 2**32


class EnfugueStableDiffusionPipeline(StableDiffusionPipeline):
    """
    This pipeline merges all of the following:
    1. txt2img
    2. img2img
    3. inpainting/outpainting
    4. controlnet
    5. tensorrt
    """

    controlnet: Optional[ControlNetModel]
    scheduler: DDIMScheduler  # Must use DDIM for multi-diffusion

    def __init__(
        self,
        vae: AutoencoderKL,
        text_encoder: CLIPTextModel,
        tokenizer: CLIPTokenizer,
        unet: UNet2DConditionModel,
        controlnet: Optional[ControlNetModel],
        scheduler: KarrasDiffusionSchedulers,
        safety_checker: StableDiffusionSafetyChecker,
        feature_extractor: CLIPImageProcessor,
        requires_safety_checker: bool = True,
        engine_size: int = 512,
        chunking_size: int = 32,
        chunking_blur: int = 64,
    ) -> None:
        super(EnfugueStableDiffusionPipeline, self).__init__(
            vae,
            text_encoder,
            tokenizer,
            unet,
            scheduler,
            safety_checker,
            feature_extractor,
            requires_safety_checker,
        )
        # Override scheduler to DDIM for multidiffusion
        self.scheduler = DDIMScheduler.from_config(self.scheduler.config)

        # Enfugue engine settings
        self.engine_size = engine_size
        self.chunking_size = chunking_size
        self.chunking_blur = chunking_blur

        # Hide tqdm
        self.set_progress_bar_config(disable=True)

        # Add an image processor for later
        self.image_processor = VaeImageProcessor(vae_scale_factor=self.vae_scale_factor)

        # Register controlnet, it can be None
        self.register_modules(controlnet=controlnet)

    @contextmanager
    def get_runtime_context(
        self, batch_size: int, device: Union[str, torch.device]
    ) -> Iterator[None]:
        """
        Used by other implementations (tensorrt), but not base.
        """
        if isinstance(device, str):
            device = torch.device(device)
        if device.type == "cpu":
            with torch.autocast("cpu"):
                yield
        else:
            yield

    def load_lora_weights(
        self,
        pretrained_model_name_or_path_or_dict: Union[str, Dict[str, torch.Tensor]],
        multiplier: int = 1,
        dtype: torch.dtype = torch.float32,
        **kwargs: Any,
    ) -> None:
        """
        Fix adapted from here: https://github.com/huggingface/diffusers/issues/3064#issuecomment-1545013909
        """
        if not isinstance(
            pretrained_model_name_or_path_or_dict, str
        ) or not pretrained_model_name_or_path_or_dict.endswith(".safetensors"):
            return super(EnfugueStableDiffusionPipeline, self).load_lora_weights(
                pretrained_model_name_or_path_or_dict, **kwargs
            )

        LORA_PREFIX_UNET = "lora_unet"
        LORA_PREFIX_TEXT_ENCODER = "lora_te"

        # load LoRA weight from .safetensors
        state_dict = safetensors.torch.load_file(
            pretrained_model_name_or_path_or_dict, device="cpu"
        )

        updates: Mapping[str, Any] = defaultdict(dict)
        for key, value in state_dict.items():
            # it is suggested to print out the key, it usually will be something like below
            # "lora_te_text_model_encoder_layers_0_self_attn_k_proj.lora_down.weight"

            layer, elem = key.split(".", 1)
            updates[layer][elem] = value

        index = 0
        # directly update weight in diffusers model
        for layer, elems in updates.items():
            index += 1

            if "text" in layer:
                layer_infos = layer.split(LORA_PREFIX_TEXT_ENCODER + "_")[-1].split("_")
                curr_layer = self.text_encoder
            else:
                layer_infos = layer.split(LORA_PREFIX_UNET + "_")[-1].split("_")
                curr_layer = self.unet

            # find the target layer
            temp_name = layer_infos.pop(0)
            while len(layer_infos) > -1:
                try:
                    curr_layer = curr_layer.__getattr__(temp_name)
                    if len(layer_infos) > 0:
                        temp_name = layer_infos.pop(0)
                    elif len(layer_infos) == 0:
                        break
                except Exception:
                    if len(temp_name) > 0:
                        temp_name += "_" + layer_infos.pop(0)
                    else:
                        temp_name = layer_infos.pop(0)

            # get elements for this layer
            weight_up = elems["lora_up.weight"].to(dtype)
            weight_down = elems["lora_down.weight"].to(dtype)
            alpha = elems["alpha"]
            if alpha:
                alpha = alpha.item() / weight_up.shape[1]
            else:
                alpha = 1.0

            # update weight
            if len(weight_up.shape) == 4:
                curr_layer.weight.data += (
                    multiplier
                    * alpha
                    * torch.mm(weight_up.squeeze(3).squeeze(2), weight_down.squeeze(3).squeeze(2))
                    .unsqueeze(2)
                    .unsqueeze(3)
                )
            else:
                curr_layer.weight.data += multiplier * alpha * torch.mm(weight_up, weight_down)

    def denormalize_latents(self, latents: torch.Tensor) -> torch.Tensor:
        """
        Denomalizes image data from [-1, 1] to [0, 1]
        """
        return (latents / 2 + 0.5).clamp(0, 1)

    @torch.no_grad()
    def prepare_mask_and_image(
        self,
        mask: Union[np.ndarray, PIL.Image.Image, torch.Tensor],
        image: Union[np.ndarray, PIL.Image.Image, torch.Tensor],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Prepares a mask and image for inpainting.
        """
        if isinstance(image, torch.Tensor):
            if not isinstance(mask, torch.Tensor):
                raise TypeError(f"`image` is a torch.Tensor but `mask` (type: {type(mask)} is not")

            # Batch single image
            if image.ndim == 3:
                assert image.shape[0] == 3, "Image outside a batch should be of shape (3, H, W)"
                image = image.unsqueeze(0)

            # Batch and add channel dim for single mask
            if mask.ndim == 2:
                mask = mask.unsqueeze(0).unsqueeze(0)

            # Batch single mask or add channel dim
            if mask.ndim == 3:
                # Single batched mask, no channel dim or single mask not batched but channel dim
                if mask.shape[0] == 1:
                    mask = mask.unsqueeze(0)

                # Batched masks no channel dim
                else:
                    mask = mask.unsqueeze(1)

            assert image.ndim == 4 and mask.ndim == 4, "Image and Mask must have 4 dimensions"
            assert (
                image.shape[-2:] == mask.shape[-2:]
            ), "Image and Mask must have the same spatial dimensions"
            assert image.shape[0] == mask.shape[0], "Image and Mask must have the same batch size"

            # Check image is in [-1, 1]
            if image.min() < -1 or image.max() > 1:
                raise ValueError("Image should be in [-1, 1] range")

            # Check mask is in [0, 1]
            if mask.min() < 0 or mask.max() > 1:
                raise ValueError("Mask should be in [0, 1] range")

            # Binarize mask
            mask[mask < 0.5] = 0
            mask[mask >= 0.5] = 1

            # Image as float32
            image = image.to(dtype=torch.float32)
        elif isinstance(mask, torch.Tensor):
            raise TypeError(f"`mask` is a torch.Tensor but `image` (type: {type(image)} is not")
        else:
            # preprocess image
            if isinstance(image, (PIL.Image.Image, np.ndarray)):
                image = [image]

            if isinstance(image, list) and isinstance(image[0], PIL.Image.Image):
                image = [np.array(i.convert("RGB"))[None, :] for i in image]
                image = np.concatenate(image, axis=0)
            elif isinstance(image, list) and isinstance(image[0], np.ndarray):
                image = np.concatenate([i[None, :] for i in image], axis=0)

            image = image.transpose(0, 3, 1, 2)
            image = torch.from_numpy(image).to(dtype=torch.float32) / 127.5 - 1.0

            # preprocess mask
            if isinstance(mask, (PIL.Image.Image, np.ndarray)):
                mask = [mask]

            if isinstance(mask, list) and isinstance(mask[0], PIL.Image.Image):
                mask = np.concatenate(
                    [np.array(m.convert("L"))[None, None, :] for m in mask], axis=0
                )
                mask = mask.astype(np.float32) / 255.0
            elif isinstance(mask, list) and isinstance(mask[0], np.ndarray):
                mask = np.concatenate([m[None, None, :] for m in mask], axis=0)

            mask[mask < 0.5] = 0
            mask[mask >= 0.5] = 1
            mask = torch.from_numpy(mask)

        masked_image = image * (mask < 0.5)

        return mask, masked_image

    @torch.no_grad()
    def create_latents(
        self,
        batch_size: int,
        num_channels_latents: int,
        height: int,
        width: int,
        dtype: torch.dtype,
        device: Union[str, torch.device],
        generator: Optional[torch.Generator] = None,
    ) -> torch.Tensor:
        """
        Creates random latents of a particular shape and type.
        """
        shape = (
            batch_size,
            num_channels_latents,
            height // self.vae_scale_factor,
            width // self.vae_scale_factor,
        )
        logger.debug(f"Creating random latents of shape {shape} and type {dtype}")
        random_latents = randn_tensor(shape, generator=generator, device=device, dtype=dtype)
        return random_latents * self.scheduler.init_noise_sigma

    @torch.no_grad()
    def encode_image_unchunked(
        self, image: torch.Tensor, generator: Optional[torch.Generator] = None
    ) -> torch.Tensor:
        """
        Encodes an image without chunking using the VAE.
        """
        return self.vae.encode(image).latent_dist.sample(generator) * self.vae.config.scaling_factor

    @torch.no_grad()
    def encode_image(
        self,
        image: torch.Tensor,
        device: Union[str, torch.device],
        generator: Optional[torch.Generator] = None,
        progress_callback: Optional[Callable[[], None]] = None,
    ) -> torch.Tensor:
        """
        Encodes an image in chunks using the VAE.
        """
        _, _, height, width = image.shape
        chunks = self.get_chunks(height, width)
        total_steps = len(chunks)

        if total_steps == 1:
            result = self.encode_image_unchunked(image, generator)
            if progress_callback is not None:
                progress_callback()
            return result

        logger.debug(f"Encoding image in {total_steps} steps.")

        latent_height = height // self.vae_scale_factor
        latent_width = width // self.vae_scale_factor

        engine_latent_size = self.engine_size // self.vae_scale_factor
        num_channels = self.vae.config.latent_channels

        count = torch.zeros((1, num_channels, latent_height, latent_width)).to(device=device)
        value = torch.zeros_like(count)

        for i, (top, bottom, left, right) in enumerate(chunks):
            top_px = top * self.vae_scale_factor
            bottom_px = bottom * self.vae_scale_factor
            left_px = left * self.vae_scale_factor
            right_px = right * self.vae_scale_factor

            image_view = image[:, :, top_px:bottom_px, left_px:right_px]

            encoded_image = self.vae.encode(image_view).latent_dist.sample(generator).to(device)

            # Blur edges as needed
            multiplier = torch.ones_like(encoded_image)
            if self.chunking_blur > 0:
                chunking_blur_latent_size = self.chunking_blur // self.vae_scale_factor
                blur_left = left > 0
                blur_top = top > 0
                blur_right = right < latent_width
                blur_bottom = bottom < latent_height

                for j in range(chunking_blur_latent_size):
                    mult = (j + 1) / (chunking_blur_latent_size + 1)
                    if blur_left:
                        multiplier[:, :, :, j] *= mult
                    if blur_top:
                        multiplier[:, :, j, :] *= mult
                    if blur_right:
                        multiplier[:, :, :, engine_latent_size - j - 1] *= mult
                    if blur_bottom:
                        multiplier[:, :, engine_latent_size - j - 1, :] *= mult

            value[:, :, top:bottom, left:right] += encoded_image * multiplier
            count[:, :, top:bottom, left:right] += multiplier

            if progress_callback is not None:
                progress_callback()

        return torch.where(count > 0, value / count, value) * self.vae.config.scaling_factor

    @torch.no_grad()
    def prepare_image_latents(
        self,
        image: Union[torch.Tensor, PIL.Image.Image],
        timestep: torch.Tensor,
        batch_size: int,
        dtype: torch.dtype,
        device: Union[str, torch.device],
        generator: Optional[torch.Generator] = None,
        progress_callback: Optional[Callable[[], None]] = None,
    ) -> torch.Tensor:
        """
        Prepares latents from an image, adding initial noise for img2img inference
        """
        image = image.to(device=device, dtype=dtype)
        init_latents = self.encode_image(
            image, device=device, generator=generator, progress_callback=progress_callback
        ).to(dtype=dtype)

        if batch_size > init_latents.shape[0] and batch_size % init_latents.shape[0] == 0:
            # duplicate images to match batch size
            additional_image_per_prompt = batch_size // init_latents.shape[0]
            init_latents = torch.cat([init_latents] * additional_image_per_prompt, dim=0)
        elif batch_size > init_latents.shape[0] and batch_size % init_latents.shape[0] != 0:
            raise ValueError(
                f"Cannot duplicate `image` of batch size {init_latents.shape[0]} to {batch_size} text prompts."
            )
        else:
            init_latents = torch.cat([init_latents], dim=0)

        shape = init_latents.shape
        noise = randn_tensor(shape, generator=generator, device=device, dtype=dtype)

        # add noise in accordance with timesteps
        return self.scheduler.add_noise(init_latents, noise, timestep)

    @torch.no_grad()
    def prepare_mask_latents(
        self,
        mask: Union[PIL.Image.Image, torch.Tensor],
        image: Union[PIL.Image.Image, torch.Tensor],
        batch_size: int,
        height: int,
        width: int,
        dtype: torch.dtype,
        device: Union[str, torch.device],
        generator: Optional[torch.Generator] = None,
        do_classifier_free_guidance: bool = False,
        progress_callback: Optional[Callable[[], None]] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Prepares both mask and image latents for inpainting
        """
        tensor_height = height // self.vae_scale_factor
        tensor_width = width // self.vae_scale_factor
        tensor_size = (tensor_height, tensor_width)
        mask = torch.nn.functional.interpolate(mask, size=tensor_size)
        mask = mask.to(device=device, dtype=dtype)
        image = image.to(device=device, dtype=dtype)

        latents = self.encode_image(
            image, device=device, generator=generator, progress_callback=progress_callback
        ).to(device=device)

        # duplicate mask and latents for each generation per prompt, using mps friendly method
        if mask.shape[0] < batch_size:
            if not batch_size % mask.shape[0] == 0:
                raise ValueError(
                    "The passed mask and the required batch size don't match. Masks are supposed to be duplicated to"
                    f" a total batch size of {batch_size}, but {mask.shape[0]} masks were passed. Make sure the number"
                    " of masks that you pass is divisible by the total requested batch size."
                )
            mask = mask.repeat(batch_size // mask.shape[0], 1, 1, 1)
        if latents.shape[0] < batch_size:
            if not batch_size % latents.shape[0] == 0:
                raise ValueError(
                    "The passed images and the required batch size don't match. Images are supposed to be duplicated"
                    f" to a total batch size of {batch_size}, but {latents.shape[0]} images were passed."
                    " Make sure the number of images that you pass is divisible by the total requested batch size."
                )
            latents = latents.repeat(batch_size // latents.shape[0], 1, 1, 1)

        mask = torch.cat([mask] * 2) if do_classifier_free_guidance else mask
        latents = torch.cat([latents] * 2) if do_classifier_free_guidance else latents

        # aligning device to prevent device errors when concating it with the latent model input
        latents = latents.to(device=device, dtype=dtype)
        return mask, latents

    @torch.no_grad()
    def get_timesteps(
        self, num_inference_steps: int, strength: float, device: str
    ) -> Tuple[torch.Tensor, int]:
        """
        Gets the original timesteps from the scheduler based on strength when doing img2img
        """
        init_timestep = min(int(num_inference_steps * strength), num_inference_steps)

        t_start = max(num_inference_steps - init_timestep, 0)
        timesteps = self.scheduler.timesteps[t_start * self.scheduler.order :]

        return timesteps, num_inference_steps - t_start

    @torch.no_grad()
    def predict_noise_residual(
        self,
        latents: torch.Tensor,
        timestep: torch.Tensor,
        embeddings: torch.Tensor,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        down_block_additional_residuals: Optional[List[torch.Tensor]] = None,
        mid_block_additional_residual: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Runs the UNet to predict noise residual.
        """
        return self.unet(
            latents,
            timestep,
            encoder_hidden_states=embeddings,
            cross_attention_kwargs=cross_attention_kwargs,
            down_block_additional_residuals=down_block_additional_residuals,
            mid_block_additional_residual=mid_block_additional_residual,
            return_dict=False,
        )[0]

    @torch.no_grad()
    def prepare_control_image(
        self,
        image: Union[torch.Tensor, PIL.Image.Image, List[PIL.Image.Image]],
        width: int,
        height: int,
        batch_size: int,
        num_images_per_prompt: int,
        device: Union[str, torch.Tensor],
        dtype: torch.dtype,
        do_classifier_free_guidance=False,
    ):
        """
        Prepares an image for controlnet conditioning.
        """
        if not isinstance(image, torch.Tensor):
            if isinstance(image, PIL.Image.Image):
                image = [image]

            if isinstance(image[0], PIL.Image.Image):
                images = []

                for image_ in image:
                    image_ = image_.convert("RGB")
                    image_ = image_.resize((width, height), resample=PIL_INTERPOLATION["lanczos"])
                    image_ = np.array(image_)
                    image_ = image_[None, :]
                    images.append(image_)

                image = images

                image = np.concatenate(image, axis=0)
                image = np.array(image).astype(np.float32) / 255.0
                image = image.transpose(0, 3, 1, 2)
                image = torch.from_numpy(image)
            elif isinstance(image[0], torch.Tensor):
                image = torch.cat(image, dim=0)

        image_batch_size = image.shape[0]

        if image_batch_size == 1:
            repeat_by = batch_size
        else:
            # image batch size is the same as prompt batch size
            repeat_by = num_images_per_prompt

        image = image.repeat_interleave(repeat_by, dim=0)
        if do_classifier_free_guidance:
            image = torch.cat([image] * 2)

        return image.to(device=device, dtype=dtype)

    @torch.no_grad()
    def prepare_controlnet_inpaint_control_image(
        self,
        image: PIL.Image.Image,
        mask: PIL.Image.Image,
        device: Union[str, torch.device],
        dtype: torch.dtype,
    ) -> torch.Tensor:
        """
        Combines the image and mask into a condition for controlnet inpainting.
        """
        image = np.array(image.convert("RGB")).astype(np.float32) / 255.0
        mask = np.array(mask.convert("L")).astype(np.float32) / 255.0

        assert (
            image.shape[0:1] == mask.shape[0:1]
        ), "image and image_mask must have the same image size"
        image[mask > 0.5] = -1.0  # set as masked pixel

        image = np.expand_dims(image, 0).transpose(0, 3, 1, 2)
        image = torch.from_numpy(image)

        return image.to(device=device, dtype=dtype)

    def get_minimal_chunks(self, height: int, width: int) -> List[Tuple[int, int, int, int]]:
        """
        Gets the minimum chunks that cover the shape in multiples of the engine width
        """
        latent_height = height // self.vae_scale_factor
        latent_width = width // self.vae_scale_factor

        latent_window_size = self.engine_size // self.vae_scale_factor
        horizontal_blocks = math.ceil(latent_width / latent_window_size)
        vertical_blocks = math.ceil(latent_height / latent_window_size)
        total_blocks = vertical_blocks * horizontal_blocks
        chunks = []

        for i in range(total_blocks):
            top = (i // horizontal_blocks) * latent_window_size
            bottom = top + latent_window_size

            left = (i % horizontal_blocks) * latent_window_size
            right = left + latent_window_size

            if bottom > latent_height:
                offset = bottom - latent_height
                bottom -= offset
                top -= offset
            if right > latent_width:
                offset = right - latent_width
                right -= offset
                left -= offset
            chunks.append((top, bottom, left, right))
        return chunks

    def get_chunks(self, height: int, width: int) -> List[Tuple[int, int, int, int]]:
        """
        Gets the chunked latent indices for multidiffusion
        """
        latent_height = height // self.vae_scale_factor
        latent_width = width // self.vae_scale_factor

        if not self.chunking_size:
            return [(0, latent_height, 0, latent_width)]

        latent_chunking_size = self.chunking_size // self.vae_scale_factor
        latent_window_size = self.engine_size // self.vae_scale_factor

        vertical_blocks = math.ceil((latent_height - latent_window_size) / latent_chunking_size + 1)
        horizontal_blocks = math.ceil(
            (latent_width - latent_window_size) / latent_chunking_size + 1
        )
        total_blocks = vertical_blocks * horizontal_blocks
        chunks = []

        for i in range(int(total_blocks)):
            top = (i // horizontal_blocks) * latent_chunking_size
            bottom = top + latent_window_size
            left = (i % horizontal_blocks) * latent_chunking_size
            right = left + latent_window_size

            if bottom > latent_height:
                offset = bottom - latent_height
                bottom -= offset
                top -= offset
            if right > latent_width:
                offset = right - latent_width
                right -= offset
                left -= offset

            chunks.append((top, bottom, left, right))

        return chunks

    def get_controlnet_conditioning_blocks(
        self,
        device: Union[str, torch.device],
        latents: torch.Tensor,
        timestep: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        controlnet_cond: Optional[torch.Tensor],
        conditioning_scale: float,
    ) -> Tuple[Optional[List[torch.Tensor]], Optional[torch.Tensor]]:
        """
        Executes the controlnet
        """
        if controlnet_cond is None or self.controlnet is None:
            return None, None

        return self.controlnet(
            latents,
            timestep,
            encoder_hidden_states=encoder_hidden_states,
            controlnet_cond=controlnet_cond,
            conditioning_scale=conditioning_scale,
            return_dict=False,
        )

    @torch.no_grad()
    def denoise_unchunked(
        self,
        height: int,
        width: int,
        device: Union[str, torch.device],
        num_inference_steps: int,
        timesteps: torch.Tensor,
        latents: torch.Tensor,
        encoded_prompt_embeds: torch.Tensor,
        guidance_scale: float,
        do_classifier_free_guidance: bool = False,
        mask: Optional[torch.Tensor] = None,
        mask_image: Optional[torch.Tensor] = None,
        control_image: Optional[torch.Tensor] = None,
        progress_callback: Optional[Callable[[], None]] = None,
        latent_callback: Optional[
            Callable[[Union[torch.Tensor, np.ndarray, List[PIL.Image.Image]]], None]
        ] = None,
        latent_callback_steps: Optional[int] = 1,
        latent_callback_type: Literal["latent", "pt", "np", "pil"] = "latent",
        conditioning_scale: float = 1.0,
        extra_step_kwargs: Optional[Dict[str, Any]] = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
    ) -> torch.Tensor:
        """
        Executes the denoising loop without chunking.
        """
        if extra_step_kwargs is None:
            extra_step_kwargs = {}

        num_steps = len(timesteps)
        num_warmup_steps = num_steps - num_inference_steps * self.scheduler.order
        logger.debug(f"Denoising image in {num_steps} steps (unchunked)")

        for i, t in enumerate(timesteps):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = (
                torch.cat([latents] * 2) if do_classifier_free_guidance else latents
            )
            latent_model_input = self.scheduler.scale_model_input(latent_model_input, t)

            # Get controlnet input if configured
            if control_image is not None:
                down_block, mid_block = self.get_controlnet_conditioning_blocks(
                    device=device,
                    latents=latent_model_input,
                    timestep=t,
                    encoder_hidden_states=encoded_prompt_embeds,
                    controlnet_cond=control_image,
                    conditioning_scale=conditioning_scale,
                )
            else:
                down_block, mid_block = None, None

            # add other dimensions to unet input if set
            if mask is not None and mask_image is not None:
                latent_model_input = torch.cat(
                    [latent_model_input, mask, mask_image],
                    dim=1,
                )

            # predict the noise residual
            noise_pred = self.predict_noise_residual(
                latent_model_input,
                t,
                encoded_prompt_embeds,
                cross_attention_kwargs,
                down_block,
                mid_block,
            )

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * (
                    noise_pred_text - noise_pred_uncond
                )

            # Compute previous noisy sample
            latents = self.scheduler.step(
                noise_pred,
                t,
                latents,
                **extra_step_kwargs,
            ).prev_sample

            if progress_callback is not None:
                progress_callback()

            # call the callback, if provided
            if (
                latent_callback is not None
                and latent_callback_steps is not None
                and i % latent_callback_steps == 0
                and (
                    i == num_steps - 1
                    or ((i + 1) > num_warmup_steps and (i + 1) % self.scheduler.order == 0)
                )
            ):
                latent_callback_value = latents

                if latent_callback_type != "latent":
                    latent_callback_value = self.decode_latents(
                        latent_callback_value, device=device, progress_callback=progress_callback
                    )
                    latent_callback_value = self.denormalize_latents(latent_callback_value)
                    if latent_callback_type != "pt":
                        latent_callback_value = self.image_processor.pt_to_numpy(
                            latent_callback_value
                        )
                        if latent_callback_type == "pil":
                            latent_callback_value = self.image_processor.numpy_to_pil(
                                latent_callback_value
                            )
                latent_callback(latent_callback_value)

        return latents

    @torch.no_grad()
    def denoise(
        self,
        height: int,
        width: int,
        device: Union[str, torch.device],
        num_inference_steps: int,
        timesteps: torch.Tensor,
        latents: torch.Tensor,
        encoded_prompt_embeds: torch.Tensor,
        guidance_scale: float,
        do_classifier_free_guidance: bool = False,
        mask: Optional[torch.Tensor] = None,
        mask_image: Optional[torch.Tensor] = None,
        control_image: Optional[torch.Tensor] = None,
        progress_callback: Optional[Callable[[], None]] = None,
        latent_callback: Optional[
            Callable[[Union[torch.Tensor, np.ndarray, List[PIL.Image.Image]]], None]
        ] = None,
        latent_callback_steps: Optional[int] = 1,
        latent_callback_type: Literal["latent", "pt", "np", "pil"] = "latent",
        conditioning_scale: float = 1.0,
        extra_step_kwargs: Optional[Dict[str, Any]] = None,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
    ) -> torch.Tensor:
        """
        Executes the denoising loop.
        """
        if extra_step_kwargs is None:
            extra_step_kwargs = {}

        chunks = self.get_chunks(height, width)
        num_chunks = len(chunks)

        if num_chunks == 1:
            return self.denoise_unchunked(
                height=height,
                width=width,
                device=device,
                num_inference_steps=num_inference_steps,
                timesteps=timesteps,
                latents=latents,
                encoded_prompt_embeds=encoded_prompt_embeds,
                guidance_scale=guidance_scale,
                do_classifier_free_guidance=do_classifier_free_guidance,
                mask=mask,
                mask_image=mask_image,
                control_image=control_image,
                conditioning_scale=conditioning_scale,
                progress_callback=progress_callback,
                latent_callback=latent_callback,
                latent_callback_steps=latent_callback_steps,
                latent_callback_type=latent_callback_type,
                extra_step_kwargs=extra_step_kwargs,
                cross_attention_kwargs=cross_attention_kwargs,
            )

        num_steps = len(timesteps)
        num_warmup_steps = num_steps - num_inference_steps * self.scheduler.order

        latent_width = width // self.vae_scale_factor
        latent_height = height // self.vae_scale_factor
        engine_latent_size = self.engine_size // self.vae_scale_factor

        count = torch.zeros_like(latents)
        value = torch.zeros_like(latents)

        total_num_steps = num_steps * num_chunks
        logger.debug(
            f"Denoising image in {total_num_steps} total steps ({num_inference_steps} inference steps * {num_chunks} chunks)"
        )

        for i, t in enumerate(timesteps):
            # zero view latents
            count.zero_()
            value.zero_()

            # iterate over chunks
            for j, (top, bottom, left, right) in enumerate(chunks):
                # Get pixel indices
                top_px = top * self.vae_scale_factor
                bottom_px = bottom * self.vae_scale_factor
                left_px = left * self.vae_scale_factor
                right_px = right * self.vae_scale_factor

                # Slice latents
                latents_for_view = latents[:, :, top:bottom, left:right]

                # expand the latents if we are doing classifier free guidance
                latent_model_input = (
                    torch.cat([latents_for_view] * 2)
                    if do_classifier_free_guidance
                    else latents_for_view
                )
                latent_model_input = self.scheduler.scale_model_input(latent_model_input, t)

                # Get controlnet input if configured
                if control_image is not None:
                    down_block, mid_block = self.get_controlnet_conditioning_blocks(
                        device=device,
                        latents=latent_model_input,
                        timestep=t,
                        encoder_hidden_states=encoded_prompt_embeds,
                        controlnet_cond=control_image[:, :, top_px:bottom_px, left_px:right_px],
                        conditioning_scale=conditioning_scale,
                    )
                else:
                    down_block, mid_block = None, None

                # add other dimensions to unet input if set
                if mask is not None and mask_image is not None:
                    latent_model_input = torch.cat(
                        [
                            latent_model_input,
                            mask[:, :, top:bottom, left:right],
                            mask_image[:, :, top:bottom, left:right],
                        ],
                        dim=1,
                    )

                # predict the noise residual
                noise_pred = self.predict_noise_residual(
                    latents=latent_model_input,
                    timestep=t,
                    embeddings=encoded_prompt_embeds,
                    cross_attention_kwargs=cross_attention_kwargs,
                    down_block_additional_residuals=down_block,
                    mid_block_additional_residual=mid_block,
                )

                # perform guidance
                if do_classifier_free_guidance:
                    noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                    noise_pred = noise_pred_uncond + guidance_scale * (
                        noise_pred_text - noise_pred_uncond
                    )

                # compute the previous noisy sample x_t -> x_t-1
                denoised_latents = self.scheduler.step(
                    noise_pred,
                    t,
                    latents_for_view,
                    **extra_step_kwargs,
                ).prev_sample

                # Blur edges as needed
                multiplier = torch.ones_like(denoised_latents)
                if self.chunking_blur > 0:
                    chunking_blur_latent_size = self.chunking_blur // self.vae_scale_factor
                    blur_left = left > 0
                    blur_top = top > 0
                    blur_right = right < latent_width
                    blur_bottom = bottom < latent_height

                    for j in range(chunking_blur_latent_size):
                        mult = (j + 1) / (chunking_blur_latent_size + 1)
                        if blur_left:
                            multiplier[:, :, :, j] *= mult
                        if blur_top:
                            multiplier[:, :, j, :] *= mult
                        if blur_right:
                            multiplier[:, :, :, engine_latent_size - j - 1] *= mult
                        if blur_bottom:
                            multiplier[:, :, engine_latent_size - j - 1, :] *= mult

                value[:, :, top:bottom, left:right] += denoised_latents * multiplier
                count[:, :, top:bottom, left:right] += multiplier

                # Call the progress callback
                if progress_callback is not None:
                    progress_callback()

            # multidiffusion
            latents = torch.where(count > 0, value / count, value)

            # call the latent_callback, if provided
            if (
                latent_callback is not None
                and latent_callback_steps is not None
                and i % latent_callback_steps == 0
                and (
                    i == num_steps - 1
                    or ((i + 1) > num_warmup_steps and (i + 1) % self.scheduler.order == 0)
                )
            ):
                latent_callback_value = latents

                if latent_callback_type != "latent":
                    latent_callback_value = self.decode_latents(
                        latent_callback_value, device=device, progress_callback=progress_callback
                    )
                    latent_callback_value = self.denormalize_latents(latent_callback_value)
                    if latent_callback_type != "pt":
                        latent_callback_value = self.image_processor.pt_to_numpy(
                            latent_callback_value
                        )
                        if latent_callback_type == "pil":
                            latent_callback_value = self.image_processor.numpy_to_pil(
                                latent_callback_value
                            )
                latent_callback(latent_callback_value)

        return latents

    @torch.no_grad()
    def decode_latent_view(self, latents: torch.Tensor) -> torch.Tensor:
        """
        Issues the command to decode a chunk of latents with the VAE.
        """
        return self.vae.decode(latents, return_dict=False)[0]

    @torch.no_grad()
    def decode_latents_unchunked(
        self, latents: torch.Tensor, device: Union[str, torch.device]
    ) -> torch.Tensor:
        """
        Decodes the latents using the VAE without chunking.
        """
        return self.decode_latent_view(latents).to(device=device)

    @torch.no_grad()
    def decode_latents(
        self,
        latents: torch.Tensor,
        device: Union[str, torch.device],
        progress_callback: Optional[Callable[[], None]] = None,
    ) -> torch.Tensor:
        """
        Decodes the latents in chunks as necessary.
        """
        samples, _, height, width = latents.shape
        height *= self.vae_scale_factor
        width *= self.vae_scale_factor

        latents = 1 / self.vae.config.scaling_factor * latents

        chunks = self.get_chunks(height, width)
        total_steps = len(chunks)

        if total_steps == 1:
            result = self.decode_latents_unchunked(latents, device)
            if progress_callback is not None:
                progress_callback()
            return result

        latent_width = width // self.vae_scale_factor
        latent_height = height // self.vae_scale_factor

        count = torch.zeros((samples, 3, height, width)).to(device=device)
        value = torch.zeros_like(count)

        logger.debug(f"Decoding latents in {total_steps} steps")

        # iterate over chunks
        for i, (top, bottom, left, right) in enumerate(chunks):
            # Slice latents
            latents_for_view = latents[:, :, top:bottom, left:right]

            # Get pixel indices
            top_px = top * self.vae_scale_factor
            bottom_px = bottom * self.vae_scale_factor
            left_px = left * self.vae_scale_factor
            right_px = right * self.vae_scale_factor

            # Decode latents
            decoded_latents = self.decode_latent_view(latents_for_view).to(device=device)

            # Blur edges as needed
            multiplier = torch.ones_like(decoded_latents)
            if self.chunking_blur > 0:
                chunking_blur_size = self.chunking_blur
                blur_left = left > 0
                blur_top = top > 0
                blur_right = right < latent_width
                blur_bottom = bottom < latent_height

                for j in range(chunking_blur_size):
                    mult = (j + 1) / (chunking_blur_size + 1)
                    if blur_left:
                        multiplier[:, :, :, j] *= mult
                    if blur_top:
                        multiplier[:, :, j, :] *= mult
                    if blur_right:
                        multiplier[:, :, :, self.engine_size - j - 1] *= mult
                    if blur_bottom:
                        multiplier[:, :, self.engine_size - j - 1, :] *= mult

            value[:, :, top_px:bottom_px, left_px:right_px] += decoded_latents * multiplier
            count[:, :, top_px:bottom_px, left_px:right_px] += multiplier

            if progress_callback is not None:
                progress_callback()

        # re-average pixels
        latents = torch.where(count > 0, value / count, value)
        return latents

    @torch.no_grad()
    def get_step_complete_callback(
        self,
        overall_steps: int,
        progress_callback: Optional[Callable[[int, int, float], None]] = None,
        log_interval: int = 10,
        log_sampling_duration: Union[int, float] = 5,
    ) -> Callable[[], None]:
        """
        Creates a scoped callback to trigger during iterations
        """
        overall_step, window_start_step, its = 0, 0, 0.0
        window_start = datetime.datetime.now()
        digits = math.ceil(math.log10(overall_steps))

        def step_complete() -> None:
            nonlocal overall_step, window_start, window_start_step, its
            overall_step += 1
            if overall_step % log_interval == 0 or overall_step == overall_steps:
                seconds_in_window = (datetime.datetime.now() - window_start).total_seconds()
                its = (overall_step - window_start_step) / seconds_in_window
                unit = "s/it" if its < 1 else "it/s"
                its = 1 / its if its < 1 else its
                logger.debug(
                    f"{{0:0{digits}d}}/{{1:0{digits}d}}: {{2:0.2f}} {{3:s}}".format(
                        overall_step, overall_steps, its, unit
                    )
                )

                if seconds_in_window > log_sampling_duration:
                    window_start_step = overall_step
                    window_start = datetime.datetime.now()

            if progress_callback is not None:
                progress_callback(overall_step, overall_steps, its)

        return step_complete

    @torch.no_grad()
    def __call__(
        self,
        prompt: Optional[Union[str, List[str]]] = None,
        image: Optional[Union[PIL.Image.Image, str]] = None,
        mask: Optional[Union[PIL.Image.Image, str]] = None,
        control_image: Optional[Union[PIL.Image.Image, str]] = None,
        height: Optional[int] = None,
        width: Optional[int] = None,
        chunking_size: Optional[int] = None,
        chunking_blur: Optional[int] = None,
        strength: float = 0.8,
        num_inference_steps: int = 50,
        guidance_scale: float = 7.5,
        conditioning_scale: float = 1.0,
        negative_prompt: Optional[Union[str, List[str]]] = None,
        num_images_per_prompt: int = 1,
        eta: float = 0.0,
        generator: Optional[torch.Generator] = None,
        latents: Optional[torch.Tensor] = None,
        prompt_embeds: Optional[torch.Tensor] = None,
        negative_prompt_embeds: Optional[torch.Tensor] = None,
        output_type: Literal["latent", "pt", "np", "pil"] = "pil",
        return_dict: bool = True,
        scale_image: bool = True,
        progress_callback: Optional[Callable[[int, int, float], None]] = None,
        latent_callback: Optional[
            Callable[[Union[torch.Tensor, np.ndarray, List[PIL.Image.Image]]], None]
        ] = None,
        latent_callback_steps: Optional[int] = 1,
        latent_callback_type: Literal["latent", "pt", "np", "pil"] = "latent",
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Union[
        StableDiffusionPipelineOutput,
        Tuple[Union[torch.Tensor, np.ndarray, List[PIL.Image.Image]], Optional[List[bool]]],
    ]:
        """
        Invokes the pipeline.
        """
        # 1. Default height and width to unet
        height = height or self.unet.config.sample_size * self.vae_scale_factor
        width = width or self.unet.config.sample_size * self.vae_scale_factor

        # Allow overridding 'chunk'
        if chunking_size is not None:
            self.chunking_size = chunking_size
        if chunking_blur is not None:
            self.chunking_blur = chunking_blur

        # Define outputs here to process later
        prepared_latents: Optional[torch.Tensor] = None
        prepared_control_image: Optional[torch.Tensor] = None
        output_nsfw: Optional[List[bool]] = None

        # 2. Define call parameters
        if prompt is not None and isinstance(prompt, str):
            batch_size = 1
        elif prompt is not None and isinstance(prompt, list):
            batch_size = len(prompt)
        elif prompt_embeds:
            batch_size = prompt_embeds.shape[0]
        else:
            raise ValueError("Prompt or prompt embeds are required.")

        if self.unet.config.in_channels == 9:
            if not image:
                logger.warning("No image present, but using inpainting model. Adding blank image.")
                image = PIL.Image.new("RGB", (width, height))
            if not mask:
                logger.warning("No mask present, but using inpainting model. Adding blank mask.")
                mask = PIL.Image.new("RGB", (width, height), (255, 255, 255))

        device = self._execution_device
        do_classifier_free_guidance = guidance_scale > 1.0

        self.scheduler.set_timesteps(num_inference_steps, device=device)
        if image and not mask:
            timesteps, num_inference_steps = self.get_timesteps(
                num_inference_steps, strength, device
            )
        else:
            timesteps = self.scheduler.timesteps

        batch_size *= num_images_per_prompt

        # Calculate chunks
        num_chunks = len(self.get_chunks(height, width))

        # Calculate total steps
        chunked_steps = 1

        if image is not None:
            chunked_steps += 1
        if mask is not None:
            chunked_steps += 1
        if latent_callback is not None and latent_callback_steps is not None:
            chunked_steps += num_inference_steps // latent_callback_steps
        overall_num_steps = num_chunks * (num_inference_steps + chunked_steps)

        step_complete = self.get_step_complete_callback(overall_num_steps, progress_callback)

        with self.get_runtime_context(batch_size, device):
            # Base runtime has no context, but extensions do
            encoded_prompt_embeds = self._encode_prompt(
                prompt,
                device,
                num_images_per_prompt,
                do_classifier_free_guidance,
                negative_prompt,
                prompt_embeds=prompt_embeds,
                negative_prompt_embeds=negative_prompt_embeds,
            )

            if image and isinstance(image, str):
                image = PIL.Image.open(image)

            if mask and isinstance(mask, str):
                mask = PIL.Image.open(mask)

            if scale_image and image:
                image_width, image_height = image.size
                if image_width != width or image_height != height:
                    logger.debug(
                        f"Resizing input image from {image_width}x{image_height} to {width}x{height}"
                    )
                    image = image.resize((width, height), resample=PIL_INTERPOLATION["lanczos"])

            if scale_image and mask:
                mask_width, mask_height = mask.size
                if mask_width != width or mask_height != height:
                    logger.debug(
                        f"Resizing input mask from {mask_width}x{mask_height} to {width}x{height}"
                    )
                    mask = mask.resize((width, height), resample=PIL_INTERPOLATION["lanczos"])

            if image:
                image = image.convert("RGB")
            if mask:
                mask = mask.convert("L")

            if image and mask:
                prepared_mask, prepared_image = self.prepare_mask_and_image(mask, image)
            elif image:
                prepared_image = self.image_processor.preprocess(image)
                prepared_mask, prepared_image_latents = None, None
            else:
                prepared_image, prepared_mask, prepared_image_latents = None, None, None

            if width < self.engine_size or height < self.engine_size:
                # Disable chunking
                logger.debug(f"{width}x{height} is smaller than is chunkable, disabling.")
                self.chunking_size = 0

            if prepared_image is not None and prepared_mask is not None:
                # Running the pipeline on an image with a mask
                num_channels_latents = self.vae.config.latent_channels

                if latents:
                    prepared_latents = latents.to(device) * self.schedule.init_noise_sigma
                else:
                    prepared_latents = self.create_latents(
                        batch_size,
                        num_channels_latents,
                        height,
                        width,
                        encoded_prompt_embeds.dtype,
                        device,
                        generator,
                    )

                prepared_mask, prepared_image_latents = self.prepare_mask_latents(
                    prepared_mask,
                    prepared_image,
                    batch_size,
                    height,
                    width,
                    encoded_prompt_embeds.dtype,
                    device,
                    generator,
                    do_classifier_free_guidance,
                    progress_callback=step_complete,
                )

                num_channels_mask = prepared_mask.shape[1]
                num_channels_masked_image = prepared_image_latents.shape[1]

                if (
                    num_channels_latents + num_channels_mask + num_channels_masked_image
                    != self.unet.config.in_channels
                ):
                    raise ValueError(
                        f"Incorrect configuration settings! The config of `pipeline.unet`: {self.unet.config} expects"
                        f" {self.unet.config.in_channels} but received `num_channels_latents`: {num_channels_latents} +"
                        f" `num_channels_mask`: {num_channels_mask} + `num_channels_masked_image`: {num_channels_masked_image}"
                        f" = {num_channels_latents+num_channels_masked_image+num_channels_mask}. Please verify the config of"
                        " `pipeline.unet` or your `mask_image` or `image` input."
                    )
            elif prepared_image is not None:
                # Running the pipeline on an image, start with that
                prepared_latents = self.prepare_image_latents(
                    prepared_image,
                    timesteps[:1].repeat(batch_size),
                    batch_size,
                    encoded_prompt_embeds.dtype,
                    device,
                    generator,
                    progress_callback=step_complete,
                )
            elif latents:
                # Running the pipeline on existing latents, add some noise
                prepared_latents = latents.to(device) * self.self.scheduler.init_noise_sigma
            else:
                # Create random latents for the whole unet
                prepared_latents = self.create_latents(
                    batch_size,
                    self.unet.config.in_channels,
                    height,
                    width,
                    encoded_prompt_embeds.dtype,
                    device,
                    generator,
                )

            # Look for controlnet and conditioning image
            if control_image is not None:
                if self.controlnet is None:
                    logger.warning("Control image passed, but no controlnet present. Ignoring.")
                    prepared_control_image = None
                else:
                    if type(control_image) is str:
                        control_image = PIL.Image.open(control_image)
                    self.controlnet = self.controlnet.to(device=device)  # Cast to device
                    prepared_control_image = self.prepare_control_image(
                        image=control_image,
                        height=height,
                        width=width,
                        batch_size=batch_size,
                        num_images_per_prompt=num_images_per_prompt,
                        device=device,
                        dtype=self.controlnet.dtype,
                        do_classifier_free_guidance=do_classifier_free_guidance,
                    )
            elif self.controlnet is not None:
                if image and mask:
                    logger.info(
                        "Assuming controlnet inpaint, creating conditioning image from image and mask"
                    )
                    prepared_control_image = self.prepare_controlnet_inpaint_control_image(
                        image=image, mask=image, device=device, dtype=self.controlnet.dtype
                    )
                else:
                    self.controlnet = self.controlnet.to("cpu")
                    logger.warning(
                        "Controlnet present, but no conditioning image. Disabling controlnet."
                    )

            # Should no longer be None
            prepared_latents = cast(torch.Tensor, prepared_latents)

            # 6. Prepare extra step kwargs.
            extra_step_kwargs = self.prepare_extra_step_kwargs(generator, eta)

            # Make sure controlnet on device
            if self.controlnet is not None:
                self.controlnet = self.controlnet.to(device=device)

            # 7. Denoising loop
            prepared_latents = self.denoise(
                height=height,
                width=width,
                device=device,
                num_inference_steps=num_inference_steps,
                timesteps=timesteps,
                latents=prepared_latents,
                encoded_prompt_embeds=encoded_prompt_embeds,
                conditioning_scale=conditioning_scale,
                guidance_scale=guidance_scale,
                do_classifier_free_guidance=do_classifier_free_guidance,
                mask=prepared_mask,
                mask_image=prepared_image_latents,
                control_image=prepared_control_image,
                progress_callback=step_complete,
                latent_callback=latent_callback,
                latent_callback_steps=latent_callback_steps,
                latent_callback_type=latent_callback_type,
                extra_step_kwargs=extra_step_kwargs,
                cross_attention_kwargs=cross_attention_kwargs,
            )

            # Clear no longer needed tensors
            del prepared_mask
            del prepared_image_latents
            del prepared_control_image

            if output_type != "latent":
                prepared_latents = self.decode_latents(
                    prepared_latents, device=device, progress_callback=step_complete
                )

        if output_type == "latent":
            output = prepared_latents
        else:
            output = self.denormalize_latents(prepared_latents)
            if output_type != "pt":
                output = self.image_processor.pt_to_numpy(output)
                output_nsfw = self.run_safety_checker(output, device, encoded_prompt_embeds.dtype)[
                    1
                ]
                if output_type == "pil":
                    output = self.image_processor.numpy_to_pil(output)

        if hasattr(self, "final_offload_hook") and self.final_offload_hook is not None:
            self.final_offload_hook.offload()

        if self.controlnet is not None:
            self.controlnet = self.controlnet.to("cpu")  # Unload controlnet from GPU

        if not return_dict:
            return (output, output_nsfw)

        return StableDiffusionPipelineOutput(images=output, nsfw_content_detected=output_nsfw)



import os
import PIL
import signal
import requests
import datetime

from typing import Any, Dict, List

from webob import Request, Response

from pibble.api.exceptions import NotFoundError
from pibble.api.server.webservice.template import TemplateServer
from pibble.api.server.webservice.jsonapi import JSONWebServiceAPIServer
from pibble.ext.user.server.base import UserExtensionHandlerRegistry
from pibble.ext.rest.server.user import UserRESTExtensionServerBase
from pibble.util.encryption import Password

from enfugue.diffusion.plan import DiffusionPlan

from enfugue.database import *
from enfugue.api.controller import *
from enfugue.interface.helpers import *

from enfugue.api.manager import SystemManager
from enfugue.api.invocations import Invocation
from enfugue.api.downloads import Download
from enfugue.api.config import EnfugueConfiguration

from enfugue.util import (
    check_make_directory,
    get_version,
    get_pending_versions,
    get_gpu_status,
    get_local_static_directory,
    logger,
)

# Auto-register some REST handlers
EnfugueAPIRESTConfiguration = {
    "root": "/api",
    "scopes": [
        {"class": "DiffusionModel", "scope": "name", "root": "models", "methods": ["GET"]},
        {"class": "User", "scope": "username", "root": "users", "methods": ["GET"]},
        {
            "class": "DiffusionInvocation",
            "scope": "id",
            "root": "invocation-history",
            "methods": ["GET"],
            "user": {"id": "user_id"},
        },
    ],
}

__all__ = ["EnfugueAPIServerBase", "EnfugueAPIServer"]


class EnfugueAPIServerBase(
    JSONWebServiceAPIServer,
    UserRESTExtensionServerBase,
):
    handlers = UserExtensionHandlerRegistry()

    @property
    def engine_root(self) -> str:
        """
        Returns the engine root location.
        """
        root = self.configuration.get("enfugue.engine.root", "~/.cache/enfugue")
        if root.startswith("~"):
            root = os.path.expanduser(root)
        root = os.path.realpath(root)
        check_make_directory(root)
        return root

    def on_configure(self) -> None:
        """
        On configuration, we add all the API objects to the ORM, create the template loader,
        the job dispatcher, and make sure we can upload files later.
        """
        self.orm.extend_base(EnfugueObjectBase)
        self.user_config = EnfugueConfiguration(self.orm)
        self.configuration.update(**self.user_config.dict())
        self.manager = SystemManager(self.configuration)
        self.manager.start_monitor()
        self.start_time = datetime.datetime.now()
        self.check_emergency_password_reset()
        if self.configuration.get("signal", True):
            self.register_signal_handlers()

    def register_signal_handlers(self) -> None:
        """
        Catch INT to make sure we destroy.
        """
        original_handler = signal.getsignal(signal.SIGINT)

        def signal_handler(*args: Any) -> None:
            self.on_destroy()
            signal.signal(signal.SIGINT, original_handler)
            raise KeyboardInterrupt()

        signal.signal(signal.SIGINT, signal_handler)

    def check_emergency_password_reset(self) -> None:
        """
        Checks if there is an emergency password reset present in the engine directory.
        """
        password_file = os.path.join(self.engine_root, "password_reset.txt")
        if os.path.exists(password_file):
            with self.orm.session() as session:
                root_user = (
                    session.query(self.orm.User).filter(self.orm.User.username == "enfugue").one()
                )
                root_user.password = Password.hash(open(password_file, "r").read())
                for token in root_user.tokens:
                    session.delete(token)
                session.commit()
            os.remove(password_file)

    def configure(self, **configuration: Any) -> None:
        """
        Override configure() to pass in REST configuration.
        """
        super(EnfugueAPIServerBase, self).configure(
            rest=EnfugueAPIRESTConfiguration, **configuration
        )

    def on_destroy(self) -> None:
        """
        Stop the manager thread when the server stops.
        """
        if hasattr(self, "manager"):
            logger.debug("Stopping system manager")
            self.manager.stop_monitor()
            self.manager.stop_engine()

    def format_plan(self, plan: DiffusionPlan) -> Dict[str, Any]:
        """
        Formats a plan for inserting into the database
        """

        def replace_images(serialized: Dict[str, Any]) -> Dict[str, Any]:
            """
            Replaces images with a metadata dictionary
            """
            for key, value in serialized.items():
                if isinstance(value, PIL.Image.Image):
                    width, height = value.size
                    metadata = {"width": width, "height": height, "mode": value.mode}
                    if hasattr(value, "filename"):
                        metadata["filename"] = value.filename
                    serialized[key] = metadata
                elif isinstance(value, dict):
                    serialized[key] = replace_images(value)
                elif isinstance(value, list):
                    serialized[key] = [
                        replace_images(part) if isinstance(part, dict) else part for part in value
                    ]
            return serialized

        return replace_images(plan.get_serialization_dict())

    def get_plan_kwargs_from_model(
        self, model_name: str, include_prompts: bool = True
    ) -> Dict[str, Any]:
        """
        Given a model name, return the keyword arguments that should be passed into a plan.
        """
        diffusion_model = (
            self.database.query(self.orm.DiffusionModel)
            .filter(self.orm.DiffusionModel.name == model_name)
            .one_or_none()
        )
        if not diffusion_model:
            raise NotFoundError(f"Unknown diffusion model {model_name}")

        model = os.path.join(self.manager.engine_root_dir, "checkpoint", diffusion_model.model)
        size = diffusion_model.size
        lora = [
            (os.path.join(self.manager.engine_root_dir, "lora", lora.model), float(lora.weight))
            for lora in diffusion_model.lora
        ]
        inversion = [
            os.path.join(self.manager.engine_root_dir, "inversion", inversion.model)
            for inversion in diffusion_model.inversion
        ]
        model_prompt = diffusion_model.prompt
        model_negative_prompt = diffusion_model.negative_prompt

        plan_kwargs: Dict[str, Any] = {
            "model": model,
            "size": size,
            "lora": lora,
            "inversion": inversion,
        }

        if include_prompts:
            plan_kwargs["model_prompt"] = model_prompt
            plan_kwargs["model_negative_prompt"] = model_negative_prompt

        return plan_kwargs

    def invoke(
        self,
        user_id: int,
        plan: DiffusionPlan,
        save: bool = True,
        disable_intermediate_decoding: bool = False,
        **kwargs: Any,
    ) -> Invocation:
        """
        Invokes the platform and saves any resulting images, returning their paths.
        """
        invocation = self.manager.invoke(
            user_id, plan, disable_intermediate_decoding=disable_intermediate_decoding, **kwargs
        )
        if save:
            self.database.add(
                self.orm.DiffusionInvocation(
                    id=invocation.uuid, user_id=user_id, plan=self.format_plan(plan)
                )
            )
            self.database.commit()
        return invocation

    def download(
        self,
        user_id: int,
        url: str,
        destination: str,
        headers: Dict[str, str] = {},
        parameters: Dict[str, Any] = {},
    ) -> Download:
        """
        Starts a download using the download manager.
        """
        return self.manager.download(
            user_id, url, destination, headers=headers, parameters=parameters
        )

    def __del__(self) -> None:
        """
        Add a __del__ in case this gets deleted before on_destroy is called
        """
        self.on_destroy()

    @handlers.path("^/api(/?)$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured()
    def status(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        Gets the status of the engine and the time since the server was started.
        """
        return {
            "status": self.manager.engine_status,
            "system": self.manager.status,
            "gpu": get_gpu_status(),
            "version": get_version(),
            "uptime": (datetime.datetime.now() - self.start_time).total_seconds(),
        }

    @handlers.path("^/api/announcements$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured()
    def announcements(self, request: Request, response: Response) -> List[Dict[str, Any]]:
        """
        Gets any announcements to display to the user.
        """
        announcements = []

        snooze_time = self.user_config.get("enfugue.snooze", None)
        snooze_duration = float("inf")
        if snooze_time is not None:
            snooze_duration = (datetime.datetime.now() - snooze_time).total_seconds()
        is_snoozed = snooze_duration < (60 * 60 * 24)
        if not is_snoozed:
            is_initialized = self.user_config.get("enfugue.initialized", False)
            if not is_initialized:
                announcements.append({"type": "initialize"})

            pending_downloads = self.manager.pending_default_downloads
            for url, dest in pending_downloads:
                model = os.path.basename(url)
                announcements.append(
                    {
                        "type": "download",
                        "url": url,
                        "size": requests.head(url, allow_redirects=True).headers["Content-Length"],
                        "model": model,
                        "destination": dest,
                    }
                )
            pending_versions = get_pending_versions()
            for version in pending_versions:
                announcements.append(
                    {
                        "type": "update",
                        "version": version["version"],
                        "release": version["release"].strftime("%Y-%m-%d"),
                        "description": version["description"],
                    }
                )

        return announcements

    @handlers.path("^/api/announcements/snooze$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured()
    def snooze_announcements(self, request: Request, response: Response) -> None:
        """
        Snoozes announcements for a day.
        """
        self.user_config["enfugue.initialized"] = True
        self.user_config["enfugue.snooze"] = datetime.datetime.now()

    @classmethod
    def serve_icon(cls, configuration: Dict[str, Any]) -> None:
        """
        When working on windows using a .exe, this runs the server using a system tray icon
        """
        import pystray
        from threading import Event

        stop_event = Event()

        def stop(icon: pystray.Icon) -> None:
            """
            Stops the server.
            """
            icon.visible = False
            stop_event.set()
            icon.stop()

        def setup(icon: pystray.Icon) -> None:
            """
            Starts the server.
            """
            icon.visible = True
            server = cls()
            server.configure_start(signal=False, **configuration)
            while not stop_event.is_set():
                stop_event.wait(1)
            server.stop()
            server.destroy()

        static_dir = get_local_static_directory()
        icon_path = configuration.get("enfugue", {}).get("icon", "favicon/favicon-64x64.png")
        icon_image = PIL.Image.open(os.path.join(static_dir, "img", icon_path))
        icon = pystray.Icon("enfugue", icon_image)
        icon.menu = pystray.Menu(pystray.MenuItem("Quit", stop))
        icon.run(setup=setup)


server_parents = tuple([EnfugueAPIServerBase] + list(EnfugueAPIControllerBase.enumerate()))

EnfugueAPIServer = type("EnfugueAPIServer", server_parents, {})



from enfugue.database.base import EnfugueObjectBase
from enfugue.database.models import DiffusionModel, DiffusionModelLora, DiffusionModelInversion
from enfugue.database.invocations import DiffusionInvocation
from enfugue.database.config import ConfigurationItem

EnfugueObjectBase, DiffusionModel, DiffusionModelLora, DiffusionModelInversion, DiffusionInvocation, ConfigurationItem

__all__ = [
    "EnfugueObjectBase",
    "ConfigurationItem",
    "DiffusionModel",
    "DiffusionModelLora",
    "DiffusionModelInversion",
    "DiffusionInvocation",
]



import os
import tempfile
from typing import Dict, Any

from enfugue.api import EnfugueAPIServer
from enfugue.interface import EnfugueInterfaceServer
from enfugue.util import get_signature, logger

__all__ = ["EnfugueSecureServer", "EnfugueServer"]


class EnfugueSecureServer(EnfugueAPIServer):
    """
    This class tries to get signing details at configuration time.
    """

    def configure(self, server=Dict[str, Any], **kwargs: Any) -> None:
        """
        At configuration, get cached or remote resources.
        """
        try:
            key, cert, chain = get_signature()
            directory = tempfile.mkdtemp()
            keyfile = os.path.join(directory, "key.pem")
            certfile = os.path.join(directory, "cert.pem")
            chainfile = os.path.join(directory, "chain.pem")
            open(keyfile, "w").write(key)
            open(certfile, "w").write(cert)
            open(chainfile, "w").write(chain)
            server["secure"] = True
            server["cert"] = certfile
            server["key"] = keyfile
            server["chain"] = chainfile
        except Exception as ex:
            logger.error(f"Couldn't get signatures, disabling SSL. {ex}")
            server["secure"] = False
        super(EnfugueSecureServer, self).configure(server=server, **kwargs)

    def on_destroy(self) -> None:
        """
        On destroy, clear tempfiles.
        """
        if hasattr(self, "_keyfile"):
            try:
                os.remove(self._keyfile)
            except:
                pass
        if hasattr(self, "_certfile"):
            try:
                os.remove(self._certfile)
            except:
                pass


class EnfugueServer(EnfugueSecureServer, EnfugueInterfaceServer):
    pass



import os
import glob
import PIL
import PIL.Image
import shutil

from typing import List, Dict, Any
from webob import Request, Response

from pibble.api.exceptions import BadRequestError, NotFoundError
from pibble.util.files import load_json
from pibble.ext.user.server.base import UserExtensionHandlerRegistry

from enfugue.api.controller.base import EnfugueAPIControllerBase
from enfugue.database.models import DiffusionModel
from enfugue.diffusion.manager import DiffusionPipelineManager
from enfugue.diffusion.plan import DiffusionPlan, DiffusionStep, DiffusionNode

__all__ = ["EnfugueAPIModelsController"]


class EnfugueAPIModelsController(EnfugueAPIControllerBase):
    handlers = UserExtensionHandlerRegistry()

    @handlers.path("^/api/checkpoints$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured()
    def get_checkpoints(self, request: Request, response: Response) -> List[str]:
        """
        Gets installed checkpoints.
        """
        checkpoints = []
        checkpoints_dir = os.path.join(self.engine_root, "checkpoint")
        if os.path.exists(checkpoints_dir):
            checkpoints = os.listdir(checkpoints_dir)
        return checkpoints

    @handlers.path("^/api/lora$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured()
    def get_lora(self, request: Request, response: Response) -> List[str]:
        """
        Gets installed lora.
        """
        lora = []
        lora_dir = os.path.join(self.engine_root, "lora")
        if os.path.exists(lora_dir):
            lora = os.listdir(lora_dir)
        return lora

    @handlers.path("^/api/inversions$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured()
    def get_inversions(self, request: Request, response: Response) -> List[str]:
        """
        Gets installed textual inversions.
        """
        inversions = []
        inversions_dir = os.path.join(self.engine_root, "inversion")
        if os.path.exists(inversions_dir):
            inversions = os.listdir(inversions_dir)
        return inversions

    @handlers.path("^/api/tensorrt$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured("DiffusionModel", "read")
    def get_tensorrt_engines(self, request: Request, response: Response) -> List[Dict[str, Any]]:
        """
        Finds engines in the model directory and determines their metadata and status.
        """
        engines = []
        for engine in glob.glob(f"{self.engine_root}/**/engine.plan", recursive=True):
            engine_dir = os.path.abspath(os.path.dirname(engine))
            engine_type = os.path.basename(os.path.dirname(engine_dir))
            engine_key = os.path.basename(os.path.dirname(engine))
            engine_model = os.path.basename(
                os.path.dirname(os.path.dirname(os.path.dirname(engine_dir)))
            )
            engine_model_name = engine_model
            engine_metadata_path = os.path.join(engine_dir, "metadata.json")
            engine_used = False
            engine_used_by = []
            engine_lora = []
            engine_inversion = []
            engine_metadata = {}
            engine_size = 512

            if os.path.exists(engine_metadata_path):
                engine_metadata = load_json(engine_metadata_path)
                engine_lora = engine_metadata.get("lora", [])
                engine_inversion = engine_metadata.get("inversion", [])

                engine_lora_dict: Dict[str, float] = dict(
                    [(str(part[0]), float(part[1])) for part in engine_lora]
                )

                engine_size = engine_metadata.get("size", engine_size)
                engine_used = False
                maybe_name, _, maybe_inpainting = engine_model.rpartition("-")
                if maybe_inpainting == "inpainting":
                    engine_model_name = maybe_name

                possible_models = (
                    self.database.query(self.orm.DiffusionModel)
                    .filter(
                        (self.orm.DiffusionModel.model == f"{engine_model_name}.ckpt")
                        | (self.orm.DiffusionModel.model == f"{engine_model_name}.safetensors")
                    )
                    .filter(self.orm.DiffusionModel.size == engine_size)
                    .all()
                )

                for model in possible_models:
                    mismatched = False
                    matched_lora = []
                    matched_inversion = []
                    for lora in model.lora:
                        lora_name, ext = os.path.splitext(lora.model)
                        if engine_lora_dict.get(lora_name, None) != lora.weight:
                            mismatched = True
                            continue
                        else:
                            matched_lora.append(lora_name)
                    for inversion in model.inversion:
                        inversion_name, ext = os.path.splitext(inversion.model)
                        if inversion_name not in engine_inversion:
                            mismatched = True
                            continue
                        else:
                            matched_inversion.append(inversion_name)
                    if (
                        len(matched_lora) == len(engine_lora_dict.keys())
                        and len(matched_inversion) == len(engine_inversion)
                        and not mismatched
                    ):
                        engine_used_by.append(model.name)
                        engine_used = True

            engines.append(
                {
                    "key": engine_key,
                    "type": engine_type,
                    "model": engine_model,
                    "lora": engine_lora,
                    "inversion": engine_inversion,
                    "used": engine_used,
                    "used_by": list(set(engine_used_by)),
                    "size": engine_size,
                    "bytes": os.path.getsize(engine),
                }
            )
        return engines

    @handlers.path(
        "^/api/tensorrt/(?P<model_name>[^\/]+)/(?P<engine_type>[a-z]+)/(?P<engine_key>[a-zA-Z0-9]+)$"
    )
    @handlers.methods("DELETE")
    @handlers.format()
    @handlers.secured("DiffusionModel", "update")
    def delete_tensorrt_engine(
        self,
        request: Request,
        response: Response,
        model_name: str,
        engine_type: str,
        engine_key: str,
    ) -> None:
        """
        Removes an individual tensorrt engine.
        """
        engine_dir = os.path.join(
            self.engine_root, "models", model_name, "tensorrt", engine_type, engine_key
        )
        if not os.path.exists(engine_dir):
            raise NotFoundError(
                f"Couldn't find {engine_type} TensorRT engine for {model_name} with key {engine_key}"
            )
        shutil.rmtree(engine_dir)

    @handlers.path("^/api/models/(?P<model_name>[^\/]+)/tensorrt$")
    @handlers.methods("GET")
    @handlers.format()
    @handlers.secured("DiffusionModel", "read")
    def get_model_tensorrt_status(
        self, request: Request, response: Response, model_name: str
    ) -> Dict[str, Any]:
        """
        Gets TensorRT status for a particular model
        """
        model = (
            self.database.query(self.orm.DiffusionModel)
            .filter(self.orm.DiffusionModel.name == model_name)
            .one_or_none()
        )
        if not model:
            raise NotFoundError(f"No model named {model_name}")

        return DiffusionPipelineManager.get_tensorrt_status(
            self.engine_root,
            model.model,
            model.size,
            [(lora.model, lora.weight) for lora in model.lora],
            [inversion.model for inversion in model.inversion],
        )

    @handlers.path("^/api/models/(?P<model_name>[^\/]+)/tensorrt/(?P<network_name>[^\/]+)$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured("DiffusionModel", "update")
    def create_model_tensorrt_engine(
        self, request: Request, response: Response, model_name: str, network_name: str
    ) -> Dict[str, Any]:
        """
        Issues a job to create an engine.
        """
        plan = DiffusionPlan(**self.get_plan_kwargs_from_model(model_name, include_prompts=False))
        plan.build_tensorrt = True

        step = DiffusionStep(
            prompt="a green field, blue sky, outside", width=plan.size, height=plan.size
        )
        network_name = network_name.lower()

        if network_name == "inpaint_unet":
            step.image = PIL.Image.new("RGB", (plan.size, plan.size))
            step.mask = PIL.Image.new("RGB", (plan.size, plan.size))
        elif network_name == "controlled_unet":
            step.control_image = PIL.Image.new("RGB", (plan.size, plan.size))
            step.controlnet = "canny"  # Have to pick one, go with the first choice
        elif network_name != "unet":
            raise BadRequestError(f"Unknown or unsupported network {network_name}")

        build_metadata = {"model": model_name, "network": network_name}
        plan.nodes = [DiffusionNode([(0, 0), (plan.size, plan.size)], step)]
        plan.image_callback_steps = None  # Disable decoding
        return self.invoke(
            request.token.user.id,
            plan,
            save=False,
            disable_intermediate_decoding=True,
            communication_timeout=3600,
            metadata={"tensorrt_build": build_metadata},
        ).format()

    @handlers.path("^/api/models/(?P<model_name>[^\/]+)$")
    @handlers.methods("PUT")
    @handlers.format()
    @handlers.secured("DiffusionModel", "update")
    def modify_model(self, request: Request, response: Response, model_name: str) -> DiffusionModel:
        """
        Asks the pipeline manager for information about models.
        """
        model = (
            self.database.query(self.orm.DiffusionModel)
            .filter(self.orm.DiffusionModel.name == model_name)
            .one_or_none()
        )

        if not model:
            raise NotFoundError(f"No model named {model_name}")

        model.name = request.parsed.get("name", model.name)
        model.model = request.parsed.get("checkpoint", model.model)
        model.size = request.parsed.get("size", model.size)
        model.prompt = request.parsed.get("prompt", model.prompt)
        model.negative_prompt = request.parsed.get("negative_prompt", model.negative_prompt)

        for existing_lora in model.lora:
            self.database.delete(existing_lora)

        for existing_inversion in model.inversion:
            self.database.delete(existing_inversion)

        self.database.commit()

        for lora in request.parsed.get("lora", []):
            new_lora = self.orm.DiffusionModelLora(
                diffusion_model_name=model.name, model=lora["model"], weight=lora["weight"]
            )
            self.database.add(new_lora)

        for inversion in request.parsed.get("inversion", []):
            new_inversion = self.orm.DiffusionModelInversion(
                diffusion_model_name=model.name,
                model=inversion,
            )
            self.database.add(new_inversion)

        self.database.commit()
        return model

    @handlers.path("^/api/models/(?P<model_name>.+)$")
    @handlers.methods("DELETE")
    @handlers.secured("DiffusionModel", "delete")
    def delete_model(self, request: Request, response: Response, model_name: str) -> None:
        """
        Asks the pipeline manager for information about models.
        """
        model = (
            self.database.query(self.orm.DiffusionModel)
            .filter(self.orm.DiffusionModel.name == model_name)
            .one_or_none()
        )
        if not model:
            raise NotFoundError(f"No model named {model_name}")

        for lora in model.lora:
            self.database.delete(lora)
        for inversion in model.inversion:
            self.database.delete(inversion)

        self.database.commit()
        self.database.delete(model)
        self.database.commit()

    @handlers.path("^/api/models$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured("DiffusionModel", "create")
    def create_model(self, request: Request, response: Response) -> DiffusionModel:
        """
        Creates a new model.
        """
        try:
            new_model = self.orm.DiffusionModel(
                name=request.parsed["name"],
                model=request.parsed["checkpoint"],
                size=request.parsed.get("size", 512),
                prompt=request.parsed.get("prompt", ""),
                negative_prompt=request.parsed.get("negative_prompt", ""),
            )
            self.database.add(new_model)
            self.database.commit()
            for lora in request.parsed.get("lora", []):
                new_lora = self.orm.DiffusionModelLora(
                    diffusion_model_name=new_model.name, model=lora["model"], weight=lora["weight"]
                )
                self.database.add(new_lora)
                self.database.commit()
            for inversion in request.parsed.get("inversion", []):
                new_inversion = self.orm.DiffusionModelInversion(
                    diffusion_model_name=new_model.name, model=inversion
                )
                self.database.add(new_inversion)
                self.database.commit()
            return new_model
        except KeyError as ex:
            raise BadRequestError(f"Missing required parameter {ex}")



from typing import Dict, Union, Iterator, Tuple

__all__ = ["TokenMerger"]


class TokenMerger:
    """
    This class allows for merging prompts in a weighted manner.
    """

    tokens: Dict[str, Union[int, float]]

    def __init__(self, *initial_phrases: str) -> None:
        self.tokens = {}
        for phrase in initial_phrases:
            self.add(phrase)

    def add(self, phrase: str, weight: Union[int, float] = 1) -> None:
        """
        Adds a token to the weights.
        """
        tokens = phrase.split(",")
        token_count = len(tokens)
        for i, token in enumerate(tokens):
            token_weight_add = min(token.count("("), token.count(")"))
            token_weight_remove = min(token.count("["), token.count("]"))
            token_weight_mult = token_weight_add - token_weight_remove
            token_immediacy_weight = max((token_count - i) / token_count, 0.5)

            if token_weight_mult < 1:
                weight /= (token_weight_mult * -1) + 1
            elif token_weight_mult > 1:
                weight *= token_weight_mult + 1
            weight *= token_immediacy_weight
            token = token.strip(" (){}[]")
            if token not in self.tokens:
                self.tokens[token] = weight
            self.tokens[token] += weight

    def __iter__(self) -> Iterator[Tuple[str, Union[int, float]]]:
        """
        Iterates over tokens in-order.
        """
        total_weight = sum(self.tokens.values())
        for key, value in reversed(sorted(self.tokens.items(), key=lambda kv: kv[1])):
            yield (key, value / total_weight)

    def __str__(self) -> str:
        """
        Stringifies the tokens.
        """
        return ",".join([token for token, weight in iter(self)])
